#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iVertexShaderId, $iFragmentShaderId, $iProgramId
Global $iVaoId, $iVboId, $iColorBufferId

Global $sVertexShader = _
	"#version 330" & @CRLF & _
	"layout(location=0) in vec4 in_Position;" & @CRLF & _
	"layout(location=1) in vec4 in_Color;" & @CRLF & _
	"out vec4 ex_Color;" & @CRLF & _
	"void main(void)" & @CRLF & _
	"{" & @CRLF & _
	"  gl_Position = in_Position;" & @CRLF & _
	"  ex_Color = in_Color;" & @CRLF & _
	"}"

Global $sFragmentShader =  _
	"#version 330" & @CRLF & _
	"in vec4 ex_Color;" & @CRLF & _
	"out vec4 out_Color;" & @CRLF & _
	"void main(void)" & @CRLF & _
	"{" & @CRLF & _
	"  out_Color = ex_Color;" & @CRLF & _
	"}"

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "..\utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 3, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not CreateShaders() Then Return 0
	If Not CreateVBO() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func CreateShaders()

	Local $iErrorCheckValue = glGetError()

	$iVertexShaderId = glCreateShader( $GL_VERTEX_SHADER )
	glShaderSource( $iVertexShaderId, 1, $sVertexShader, StringLen( $sVertexShader ) )
	glCompileShader( $iVertexShaderId )

	$iFragmentShaderId = glCreateShader( $GL_FRAGMENT_SHADER )
	glShaderSource( $iFragmentShaderId, 1, $sFragmentShader, StringLen( $sFragmentShader ) )
	glCompileShader( $iFragmentShaderId )

	$iProgramId = glCreateProgram()
	glAttachShader( $iProgramId, $iVertexShaderId )
	glAttachShader( $iProgramId, $iFragmentShaderId )
	glLinkProgram( $iProgramId )
	glUseProgram( $iProgramId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not create the shaders: " & _
			gluErrorString( $iErrorCheckValue ) )
		Return 0
	EndIf

	Return 1

EndFunc


Func DestroyShaders()

	Local $iErrorCheckValue = glGetError()

	glUseProgram( 0 )

	glDetachShader( $iProgramId, $iVertexShaderId )
	glDetachShader( $iProgramId, $iFragmentShaderId )

	glDeleteShader( $iFragmentShaderId )
	glDeleteShader( $iVertexShaderId )

	glDeleteProgram( $iProgramId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not destroy the shaders: " & _
			gluErrorString( $iErrorCheckValue ) )
	EndIf

EndFunc


Func CreateVBO()

	Local $aVertices[16] = [ _
		-0.8,  0.8, 0.0, 1.0, _
		 0.8,  0.8, 0.0, 1.0, _
		-0.8, -0.8, 0.0, 1.0, _
		 0.8, -0.8, 0.0, 1.0 ]

	Local $aColors[16] = [ _
		1.0, 0.0, 0.0, 1.0, _
		0.0, 1.0, 0.0, 1.0, _
		0.0, 0.0, 1.0, 1.0, _
		1.0, 1.0, 1.0, 1.0 ]

	Local $iErrorCheckValue = glGetError()

	glGenVertexArrays( 1, $iVaoId )
	glBindVertexArray( $iVaoId )

	glGenBuffers( 1, $iVboId )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVboId )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertices), $aVertices, $GL_STATIC_DRAW )
	glVertexAttribPointer( 0, 4, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glEnableVertexAttribArray(0)

	glGenBuffers( 1, $iColorBufferId )
	glBindBuffer( $GL_ARRAY_BUFFER, $iColorBufferId )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aColors), $aColors, $GL_STATIC_DRAW )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glEnableVertexAttribArray(1)

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not create a VBO: " & _
			gluErrorString( $iErrorCheckValue ) )
		Return 0
	EndIf

	Return 1

EndFunc


Func DestroyVBO()

	Local $iErrorCheckValue = glGetError()

	glDisableVertexAttribArray(1)
	glDisableVertexAttribArray(0)

	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glDeleteBuffers( 1, $iColorBufferId )
	glDeleteBuffers( 1, $iVboId )

	glBindVertexArray(0)
	glDeleteVertexArrays( 1, $iVaoId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not destroy the VBO: " & _
			gluErrorString( $iErrorCheckValue ) )
	EndIf

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glDrawArrays( $GL_TRIANGLE_STRIP, 0, 4 )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	DestroyVBO()
	DestroyShaders()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
