; OpenGl EXTENSION functions from glew.h

; --- References ---

; OpenGL 2.1: http://www.opengl.org/sdk/docs/man2/
; OpenGL 3.3: http://www.opengl.org/sdk/docs/man3/

; --- A note about extension functions ---

; The extension functions below are sorted by OpenGL version, extension name and function name. It should be easy to figure out which
; version and extension a function belongs to, in case you get errors when you use the function.

; glGenerateMipmap() is used as an example. It was released in the GL_ARB_framebuffer_object extension to OpenGL 2.1. If your video card
; supports 2.1, but you get errors when you use the function, it's probably because you don't have the GL_ARB_framebuffer_object extension.
; Update the drivers for your video card to get the extension.

; You can't find glGenerateMipmap() in the OpenGL 2.1 reference documentation above. But you can find it the 3.3 documentation. Somewhere
; between 2.1 and 3.3 glGenerateMipmap() has become part of the core functions. Or the extension has been upgraded to a core extension. If
; your video card supports 3.3 then this function was installed at your system, at the time you installed the video card, no matter if the
; function is a core function or belongs to a core extension.

; --- Error checking and error codes ---

; According to references most functions where the name ends on a type specification and a "v" ( <function>{bv,dv,fv,iv,sv,ubv,uiv,usv} )
; accepts a pointer to an array as a parameter. In translation to AutoIt these functions can accept both a pointer and an array. See
; glColor function below as an example. If the parameter isn't a pointer or an array an error value of 1 is returned ( @error = 1 ).
; This is implemented when the parameter is an input parameter.

; For functions that accept an array as a parameter the array is checked for size and dimension. In case of an error an error value
; of 1 is returned ( @error = 1 ).

; If wglGetProcAddress fails an error value of 2 is returned ( @error = 2 ) and the extended error value ( @extended ) is set to the
; error code returned by GetLastError.

; If a DllCall fails an error value of 3 is returned ( @error = 3 ).


#include-once

; ------------------------------ OpenGL 2.1 ------------------------------

; ---------------------- GL_ARB_framebuffer_object -----------------------

Func glGenerateMipmap( $target )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGenerateMipmap" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $target )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ---------------------- GL_ARB_vertex_array_object ----------------------

Func glBindVertexArray( $array )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glBindVertexArray" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $array )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDeleteVertexArrays( $n, ByRef $arrays )
	Local $a = 0, $array
	If IsArray( $arrays ) Then
		$a = 1
		If UBound( $arrays ) <> $n Or _
			UBound( $arrays, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDeleteVertexArrays" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			$array = $arrays[$i]
			DllCallAddress( "none", $ptr, "int", 1, "uint*", $array )
			If @error Then Return SetError( 3, 0, 0 )
		Next
	Else
		DllCallAddress( "none", $ptr, "int", 1, "uint*", $arrays )
		If @error Then Return SetError( 3, 0, 0 )
	EndIf
EndFunc

Func glGenVertexArrays( $n, ByRef $arrays )
	Local $a = 0, $array
	If IsArray( $arrays ) Then
		$a = 1
		If UBound( $arrays ) <> $n Or _
			UBound( $arrays, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGenVertexArrays" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			Local $aRet = DllCallAddress( "none", $ptr, "int", 1, "uint*", $array )
			If @error Then Return SetError( 3, 0, 0 )
			$arrays[$i] = $aRet[2]
		Next
	Else
		Local $aRet = DllCallAddress( "none", $ptr, "int", 1, "uint*", $array )
		If @error Then Return SetError( 3, 0, 0 )
		$arrays = $aRet[2]
	EndIf
EndFunc

; ------------------------------ OpenGL 3.0 ------------------------------


; ------------------------------ OpenGL 3.1 ------------------------------

; ------------------- GL_ARB_draw_elements_base_vertex -------------------

Func glDrawElementsBaseVertex( $mode, $count, $type, $indices, $basevertex )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDrawElementsBaseVertex" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $mode, "int", $count, "uint", $type, "ptr", $indices, "int", $basevertex )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; --------------------- GL_ARB_uniform_buffer_object ---------------------

Func glBindBufferRange( $target, $index, $buffer, $offset, $size )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glBindBufferRange" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $target, "uint", $index, "uint", $buffer, "int", $offset, "int", $size )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glGetUniformBlockIndex( $program, $uniformBlockName )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetUniformBlockIndex" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "int", $ptr, "uint", $program, "str", $uniformBlockName )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glUniformBlockBinding( $program, $uniformBlockIndex, $uniformBlockBinding )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniformBlockBinding" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $program, "uint", $uniformBlockIndex, "uint", $uniformBlockBinding )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ------------------------------ OpenGL 3.2 ------------------------------


; ------------------------------ OpenGL 3.3 ------------------------------


; ------------------------------ OpenGL 4.0 ------------------------------


; ------------------------------ OpenGL 4.1 ------------------------------


; ------------------------------ OpenGL 4.2 ------------------------------
