; GLM constants and functions

#include-once

#include "..\Includes\MathFuncs.au3"

; ----------------------------------- A note about OpenGL matrices -----------------------------------

; OpenGL accesses matrices in column major order, i.e. the matrix elements have the following indices:

;		 0  4  8  12
;		 1  5  9  13
;		 2  6 10  14
;		 3  7 11  15

; In AutoIt and C/C++ multidimensional arrays the row and column indices are transposed like this:

;		 0  1  2  3
;		 4  5  6  7
;		 8  9 10 11
;		12 13 14 15

; Correlation between indices in one and two dimensional AutoIt and C/C++ arrays:

;		m1[16] ~ m2[4][4] ~ [ m2[0], m2[1], m2[2], m2[3] ] ~ [ $vec4, $vec4, $vec4, $vec4 ]

;		m1[0]  = m2[0][0]	~	m2[0] ~ [ m2[0][0], m2[0][1], m2[0][2], m2[0][3] ] ~ $vec4
;		m1[1]  = m2[0][1]
;		m1[2]  = m2[0][2]
;		m1[3]  = m2[0][3]

;		m1[4]  = m2[1][0]	~	m2[1] ~ [ m2[1][0], m2[1][1], m2[1][2], m2[1][3] ] ~ $vec4
;		m1[5]  = m2[1][1]
;		m1[6]  = m2[1][2]
;		m1[7]  = m2[1][3]

;		m1[8]  = m2[2][0]	~	m2[2] ~ [ m2[2][0], m2[2][1], m2[2][2], m2[2][3] ] ~ $vec4
;		m1[9]  = m2[2][1]
;		m1[10] = m2[2][2]
;		m1[11] = m2[2][3]

;		m1[12] = m2[3][0]	~	m2[3] ~ [ m2[3][0], m2[3][1], m2[3][2], m2[3][3] ] ~ $vec4
;		m1[13] = m2[3][1]
;		m1[14] = m2[3][2]
;		m1[15] = m2[3][3]


; --------------------- 4x4 matrix functions, all functions returns a 4x4 matrix ---------------------

Global Const $NULL_MATRIX = glmInit( 0 )
Global Const $IDENTITY_MATRIX = glmInit( 1 )


Func glmInit( $val ) ; ( $num )
	Local $mat4[16] = [ _
		$val, 0, 0, 0, _
		0, $val, 0, 0, _
		0, 0, $val, 0, _
		0, 0, 0, $val ]
	Return $mat4
EndFunc


; Returns the inverse of a 4x4 matrix
Func glmInverse( ByRef $m ) ; ( $mat4 )
	; glm\core\func_matrix.hpp line 92
	; glm\core\func_matrix.inl line 506

	Local $Coef00 = $m[10] * $m[15] - $m[14] * $m[11]
	Local $Coef02 = $m[6]  * $m[15] - $m[14] * $m[7]
	Local $Coef03 = $m[6]  * $m[11] - $m[10] * $m[7]

	Local $Coef04 = $m[9]  * $m[15] - $m[13] * $m[11]
	Local $Coef06 = $m[5]  * $m[15] - $m[13] * $m[7]
	Local $Coef07 = $m[5]  * $m[11] - $m[9]  * $m[7]

	Local $Coef08 = $m[9]  * $m[14] - $m[13] * $m[10]
	Local $Coef10 = $m[5]  * $m[14] - $m[13] * $m[6]
	Local $Coef11 = $m[5]  * $m[10] - $m[9]  * $m[6]

	Local $Coef12 = $m[8]  * $m[15] - $m[12] * $m[11]
	Local $Coef14 = $m[4]  * $m[15] - $m[12] * $m[7]
	Local $Coef15 = $m[4]  * $m[11] - $m[8]  * $m[7]

	Local $Coef16 = $m[8]  * $m[14] - $m[12] * $m[10]
	Local $Coef18 = $m[4]  * $m[14] - $m[12] * $m[6]
	Local $Coef19 = $m[4]  * $m[10] - $m[8]  * $m[6]

	Local $Coef20 = $m[8]  * $m[13] - $m[12] * $m[9]
	Local $Coef22 = $m[4]  * $m[13] - $m[12] * $m[5]
	Local $Coef23 = $m[4]  * $m[9]  - $m[8]  * $m[5]

	Local $SignA[4] = [ +1, -1, +1, -1 ]
	Local $SignB[4] = [ -1, +1, -1, +1 ]

	Local $Fac0[4] = [ $Coef00, $Coef00, $Coef02, $Coef03 ]
	Local $Fac1[4] = [ $Coef04, $Coef04, $Coef06, $Coef07 ]
	Local $Fac2[4] = [ $Coef08, $Coef08, $Coef10, $Coef11 ]
	Local $Fac3[4] = [ $Coef12, $Coef12, $Coef14, $Coef15 ]
	Local $Fac4[4] = [ $Coef16, $Coef16, $Coef18, $Coef19 ]
	Local $Fac5[4] = [ $Coef20, $Coef20, $Coef22, $Coef23 ]

	Local $Vec0[4] = [ $m[4], $m[0], $m[0], $m[0] ]
	Local $Vec1[4] = [ $m[5], $m[1], $m[1], $m[1] ]
	Local $Vec2[4] = [ $m[6], $m[2], $m[2], $m[2] ]
	Local $Vec3[4] = [ $m[7], $m[3], $m[3], $m[3] ]

	;tvec4<T> Inv0 =    SignA    * (  Vec1    *  Fac0    -  Vec2    *  Fac1    +  Vec3    *  Fac2 )
	Local $Inv0[4] = [ $SignA[0] * ( $Vec1[0] * $Fac0[0] - $Vec2[0] * $Fac1[0] + $Vec3[0] * $Fac2[0] ), _
                     $SignA[1] * ( $Vec1[1] * $Fac0[1] - $Vec2[1] * $Fac1[1] + $Vec3[1] * $Fac2[1] ), _
										 $SignA[2] * ( $Vec1[2] * $Fac0[2] - $Vec2[2] * $Fac1[2] + $Vec3[2] * $Fac2[2] ), _
										 $SignA[3] * ( $Vec1[3] * $Fac0[3] - $Vec2[3] * $Fac1[3] + $Vec3[3] * $Fac2[3] ) ]

	;tvec4<T> Inv1 =    SignB    * (  Vec0    *  Fac0    -  Vec2    *  Fac3    +  Vec3    *  Fac4 )
	Local $Inv1[4] = [ $SignB[0] * ( $Vec0[0] * $Fac0[0] - $Vec2[0] * $Fac3[0] + $Vec3[0] * $Fac4[0] ), _
                     $SignB[1] * ( $Vec0[1] * $Fac0[1] - $Vec2[1] * $Fac3[1] + $Vec3[1] * $Fac4[1] ), _
										 $SignB[2] * ( $Vec0[2] * $Fac0[2] - $Vec2[2] * $Fac3[2] + $Vec3[2] * $Fac4[2] ), _
										 $SignB[3] * ( $Vec0[3] * $Fac0[3] - $Vec2[3] * $Fac3[3] + $Vec3[3] * $Fac4[3] ) ]

	;tvec4<T> Inv2 =    SignA    * (  Vec0    *  Fac1    -  Vec1    *  Fac3    +  Vec3    *  Fac5 )
	Local $Inv2[4] = [ $SignA[0] * ( $Vec0[0] * $Fac1[0] - $Vec1[0] * $Fac3[0] + $Vec3[0] * $Fac5[0] ), _
                     $SignA[1] * ( $Vec0[1] * $Fac1[1] - $Vec1[1] * $Fac3[1] + $Vec3[1] * $Fac5[1] ), _
										 $SignA[2] * ( $Vec0[2] * $Fac1[2] - $Vec1[2] * $Fac3[2] + $Vec3[2] * $Fac5[2] ), _
										 $SignA[3] * ( $Vec0[3] * $Fac1[3] - $Vec1[3] * $Fac3[3] + $Vec3[3] * $Fac5[3] ) ]

	;tvec4<T> Inv3 =    SignB    * (  Vec0    *  Fac2    -  Vec1    *  Fac4    +  Vec2    *  Fac5 )
	Local $Inv3[4] = [ $SignB[0] * ( $Vec0[0] * $Fac2[0] - $Vec1[0] * $Fac4[0] + $Vec2[0] * $Fac5[0] ), _
                     $SignB[1] * ( $Vec0[1] * $Fac2[1] - $Vec1[1] * $Fac4[1] + $Vec2[1] * $Fac5[1] ), _
										 $SignB[2] * ( $Vec0[2] * $Fac2[2] - $Vec1[2] * $Fac4[2] + $Vec2[2] * $Fac5[2] ), _
										 $SignB[3] * ( $Vec0[3] * $Fac2[3] - $Vec1[3] * $Fac4[3] + $Vec2[3] * $Fac5[3] ) ]

	;tmat4x4<T> Inverse(Inv0, Inv1, Inv2, Inv3)
	Local $Inverse[16] = [ $Inv0[0], $Inv1[0], $Inv2[0], $Inv3[0], _
	                       $Inv0[1], $Inv1[1], $Inv2[1], $Inv3[1], _
												 $Inv0[2], $Inv1[2], $Inv2[2], $Inv3[2], _
												 $Inv0[3], $Inv1[3], $Inv2[3], $Inv3[3] ]

	;tvec4<T> Row0(Inverse[0][0], Inverse[1][0], Inverse[2][0], Inverse[3][0])
	Local $Row0[4] = [ $Inverse[0], $Inverse[4], $Inverse[8], $Inverse[12] ]

	Local $m0[4] = [ $m[0], $m[4], $m[8], $m[12] ]

	Local $d = glmDotVec4Num( $m0, $Row0 )

	For $i = 0 To 15
		$Inverse[$i] /= $d
	Next

	Return $Inverse
EndFunc


; Build a look at view matrix
Func glmLookAt( ByRef $eye, ByRef $center, ByRef $up ) ; ( $vec3, $vec3, $vec3 )
	; glm\gtc\matrix_transform.hpp line 141
	; glm\gtc\matrix_transform.inl line 365

	Local $f = glmNormalizeVec3( glmSubtractVec3( $center, $eye ) )
	Local $u = glmNormalizeVec3( $up )
	Local $s = glmNormalizeVec3( glmCrossVec3( $f, $u ) )
	$u = glmCrossVec3( $s, $f )

	Local $result = $IDENTITY_MATRIX
	$result[0]  =  $s[0]
	$result[4]  =  $s[1]
	$result[8]  =  $s[2]
	$result[1]  =  $u[0]
	$result[5]  =  $u[1]
	$result[9]  =  $u[2]
	$result[2]  = -$f[0]
	$result[6]  = -$f[1]
	$result[10] = -$f[2]

	Local $vec3[3] = [ -$eye[0], -$eye[1], -$eye[2] ]
	Return glmTranslate( $result, $vec3 )
EndFunc


; Multiplies two 4x4 matrices and returns a 4x4 matrix
Func glmMultiply( ByRef $m1, ByRef $m2 ) ; ( $mat4, $mat4 )
	; glm\core\type_mat4x4.hpp line 202
	; glm\core\type_mat4x4.inl line 731

	Local $m[16]

 ;Result[0] = SrcA0 * SrcB0[0] + SrcA1 * SrcB0[1] + SrcA2 * SrcB0[2] + SrcA3 * SrcB0[3];
 ; R[0]  =   A0   * B0[0]   +  A1    * B0[1]   +  A2     * B0[2]   +  A3     * B0[3];
	$m[0]  = $m1[0] * $m2[0]  + $m1[4] * $m2[1]  + $m1[8]  * $m2[2]  + $m1[12] * $m2[3]
	$m[1]  = $m1[1] * $m2[0]  + $m1[5] * $m2[1]  + $m1[9]  * $m2[2]  + $m1[13] * $m2[3]
	$m[2]  = $m1[2] * $m2[0]  + $m1[6] * $m2[1]  + $m1[10] * $m2[2]  + $m1[14] * $m2[3]
	$m[3]  = $m1[3] * $m2[0]  + $m1[7] * $m2[1]  + $m1[11] * $m2[2]  + $m1[15] * $m2[3]

 ;Result[1] = SrcA0 * SrcB1[0] + SrcA1 * SrcB1[1] + SrcA2 * SrcB1[2] + SrcA3 * SrcB1[3];
 ; R[1]  =   A0   * B1[0]   +  A1    * B1[1]   +  A2     * B1[2]   +  A3     * B1[3];
	$m[4]  = $m1[0] * $m2[4]  + $m1[4] * $m2[5]  + $m1[8]  * $m2[6]  + $m1[12] * $m2[7]
	$m[5]  = $m1[1] * $m2[4]  + $m1[5] * $m2[5]  + $m1[9]  * $m2[6]  + $m1[13] * $m2[7]
	$m[6]  = $m1[2] * $m2[4]  + $m1[6] * $m2[5]  + $m1[10] * $m2[6]  + $m1[14] * $m2[7]
	$m[7]  = $m1[3] * $m2[4]  + $m1[7] * $m2[5]  + $m1[11] * $m2[6]  + $m1[15] * $m2[7]

 ;Result[2] = SrcA0 * SrcB2[0] + SrcA1 * SrcB2[1] + SrcA2 * SrcB2[2] + SrcA3 * SrcB2[3];
 ; R[2]  =   A0   * B2[0]   +  A1    * B2[1]   +  A2     * B2[2]   +  A3     * B2[3];
	$m[8]  = $m1[0] * $m2[8]  + $m1[4] * $m2[9]  + $m1[8]  * $m2[10] + $m1[12] * $m2[11]
	$m[9]  = $m1[1] * $m2[8]  + $m1[5] * $m2[9]  + $m1[9]  * $m2[10] + $m1[13] * $m2[11]
	$m[10] = $m1[2] * $m2[8]  + $m1[6] * $m2[9]  + $m1[10] * $m2[10] + $m1[14] * $m2[11]
	$m[11] = $m1[3] * $m2[8]  + $m1[7] * $m2[9]  + $m1[11] * $m2[10] + $m1[15] * $m2[11]

 ;Result[3] = SrcA0 * SrcB3[0] + SrcA1 * SrcB3[1] + SrcA2 * SrcB3[2] + SrcA3 * SrcB3[3];
 ; R[3]  =   A0   * B3[0]   +  A1    * B3[1]   +  A2     * B3[2]   +  A3     * B3[3];
	$m[12] = $m1[0] * $m2[12] + $m1[4] * $m2[13] + $m1[8]  * $m2[14] + $m1[12] * $m2[15]
	$m[13] = $m1[1] * $m2[12] + $m1[5] * $m2[13] + $m1[9]  * $m2[14] + $m1[13] * $m2[15]
	$m[14] = $m1[2] * $m2[12] + $m1[6] * $m2[13] + $m1[10] * $m2[14] + $m1[14] * $m2[15]
	$m[15] = $m1[3] * $m2[12] + $m1[7] * $m2[13] + $m1[11] * $m2[14] + $m1[15] * $m2[15]

	Return $m
EndFunc


; Creates a matrix for a symmetric perspective-view frustum
Func glmPerspective( $fovy, $aspect, $zNear, $zFar ) ; ( $degrees, $num, $num, $num )
	; glm\gtc\matrix_transform.hpp line 84
	; glm\gtc\matrix_transform.inl line 205

	Local $range  = Tan( Radians( $fovy / 2 ) ) * $zNear
	Local $left   = -$range * $aspect
	Local $right  = -$left
	Local $bottom = -$range
	Local $top    = $range

	Local $result = $NULL_MATRIX
	$result[0]  = ( 2 * $zNear ) / ( $right - $left )
	$result[5]  = ( 2 * $zNear ) / ( $top - $bottom )
	$result[10] = - ( $zFar + $zNear ) / ( $zFar - $zNear )
	$result[11] = - 1
	$result[14] = - ( 2 * $zFar * $zNear ) / ( $zFar - $zNear )

	Return $result
EndFunc


; Builds a rotation 4 * 4 matrix created from an axis vector and an angle expressed in degrees
Func glmRotate( ByRef $mat4, $fAngle, ByRef $axis ) ; ( $mat4, $degrees, $vec3 )
	; glm\gtc\matrix_transform.hpp line 38
	; glm\gtc\matrix_transform.inl line 26

	Local $a = Radians( $fAngle )
	Local $c = Cos( $a ), $d = 1 - $c
	Local $s = Sin( $a )
	Local $axis2 = glmNormalizeVec3( $axis )
	Local $temp[3] = [ $d*$axis2[0], $d*$axis2[1], $d*$axis2[2] ]

	Local $rotate[16], $result[16]

	$rotate[0]  = $c + $temp[0] * $axis2[0]
	$rotate[1]  = 0  + $temp[0] * $axis2[1] + $s * $axis2[2]
	$rotate[2]  = 0  + $temp[0] * $axis2[2] - $s * $axis2[1]

	$rotate[4]  = 0  + $temp[1] * $axis2[0] - $s * $axis2[2]
	$rotate[5]  = $c + $temp[1] * $axis2[1]
	$rotate[6]  = 0  + $temp[1] * $axis2[2] + $s * $axis2[0]

	$rotate[8]  = 0  + $temp[2] * $axis2[0] + $s * $axis2[1]
	$rotate[9]  = 0  + $temp[2] * $axis2[1] - $s * $axis2[0]
	$rotate[10] = $c + $temp[2] * $axis2[2]

	;Result[0]  =   m[0]   * Rotate[0][0] + m[1]   * Rotate[0][1] + m[2]    * Rotate[0][2];
	$result[0]  = $mat4[0] * $rotate[0] + $mat4[4] * $rotate[1] + $mat4[8]  * $rotate[2]
	$result[1]  = $mat4[1] * $rotate[0] + $mat4[5] * $rotate[1] + $mat4[9]  * $rotate[2]
	$result[2]  = $mat4[2] * $rotate[0] + $mat4[6] * $rotate[1] + $mat4[10] * $rotate[2]
	$result[3]  = $mat4[3] * $rotate[0] + $mat4[7] * $rotate[1] + $mat4[11] * $rotate[2]

	;Result[1]  =   m[0]   * Rotate[1][0] + m[1]   * Rotate[1][1] + m[2]    * Rotate[1][2];
	$result[4]  = $mat4[0] * $rotate[4] + $mat4[4] * $rotate[5] + $mat4[8]  * $rotate[6]
	$result[5]  = $mat4[1] * $rotate[4] + $mat4[5] * $rotate[5] + $mat4[9]  * $rotate[6]
	$result[6]  = $mat4[2] * $rotate[4] + $mat4[6] * $rotate[5] + $mat4[10] * $rotate[6]
	$result[7]  = $mat4[3] * $rotate[4] + $mat4[7] * $rotate[5] + $mat4[11] * $rotate[6]

	;Result[2]  =   m[0]   * Rotate[2][0] + m[1]   * Rotate[2][1] + m[2]    * Rotate[2][2];
	$result[8]  = $mat4[0] * $rotate[8] + $mat4[4] * $rotate[9] + $mat4[8]  * $rotate[10]
	$result[9]  = $mat4[1] * $rotate[8] + $mat4[5] * $rotate[9] + $mat4[9]  * $rotate[10]
	$result[10] = $mat4[2] * $rotate[8] + $mat4[6] * $rotate[9] + $mat4[10] * $rotate[10]
	$result[11] = $mat4[3] * $rotate[8] + $mat4[7] * $rotate[9] + $mat4[11] * $rotate[10]

	;Result[3] = m[3];
	$result[12] = $mat4[12]
	$result[13] = $mat4[13]
	$result[14] = $mat4[14]
	$result[15] = $mat4[15]

	Return $result
EndFunc


; Builds a scale 4 * 4 matrix created from 3 scalars
Func glmScale( ByRef $mat4, ByRef $vec3 )
	; glm\gtc\matrix_transform.hpp line 46
	; glm\gtc\matrix_transform.inl line 63

	Local $result[16]

	;Result[0]  =  m[0]     *  v[0];
	$result[0]  = $mat4[0]  * $vec3[0]
	$result[1]  = $mat4[1]  * $vec3[0]
	$result[2]  = $mat4[2]  * $vec3[0]
	$result[3]  = $mat4[3]  * $vec3[0]

	;Result[1]  =  m[1]     *  v[1];
	$result[4]  = $mat4[4]  * $vec3[1]
	$result[5]  = $mat4[5]  * $vec3[1]
	$result[6]  = $mat4[6]  * $vec3[1]
	$result[7]  = $mat4[7]  * $vec3[1]

	;Result[2]  =  m[2]     *  v[2];
	$result[8]  = $mat4[8]  * $vec3[2]
	$result[9]  = $mat4[9]  * $vec3[2]
	$result[10] = $mat4[10] * $vec3[2]
	$result[11] = $mat4[11] * $vec3[2]

	;Result[3]  =  m[3];
	$result[12] = $mat4[12]
	$result[13] = $mat4[13]
	$result[14] = $mat4[14]
	$result[15] = $mat4[15]

	Return $result
EndFunc


; Builds a translation 4 * 4 matrix created from a vector of 3 components
Func glmTranslate( ByRef $mat4, ByRef $vec3 )
	; glm\gtc\matrix_transform.hpp line 31
	; glm\gtc\matrix_transform.inl line 14

	Local $result = $mat4

	;Result[3]  =  m[0]    *  v[0]    +  m[1]    *  v[1]    +  m[2]     *  v[2]    +  m[3];
	$result[12] = $mat4[0] * $vec3[0] + $mat4[4] * $vec3[1] + $mat4[8]  * $vec3[2] + $mat4[12]
	$result[13] = $mat4[1] * $vec3[0] + $mat4[5] * $vec3[1] + $mat4[9]  * $vec3[2] + $mat4[13]
	$result[14] = $mat4[2] * $vec3[0] + $mat4[6] * $vec3[1] + $mat4[10] * $vec3[2] + $mat4[14]
	$result[15] = $mat4[3] * $vec3[0] + $mat4[7] * $vec3[1] + $mat4[11] * $vec3[2] + $mat4[15]

	Return $result
EndFunc


; --------------------- 3x3 matrix functions, all functions returns a 3x3 matrix ---------------------

;		m1[0] = m2[0][0]
;		m1[1] = m2[0][1]
;		m1[2] = m2[0][2]

;		m1[3] = m2[1][0]
;		m1[4] = m2[1][1]
;		m1[5] = m2[1][2]

;		m1[6] = m2[2][0]
;		m1[7] = m2[2][1]
;		m1[8] = m2[2][2]


; Returns the inverse of a 3x3 matrix
Func glmInverseMat3( ByRef $m ) ; ( $mat3 )
	; glm\core\func_matrix.hpp line 84
	; glm\core\func_matrix.inl line 477

	Local $d = $m[0] * ( $m[4] * $m[8] - $m[7] * $m[5] ) _
	         - $m[3] * ( $m[1] * $m[8] - $m[7] * $m[2] ) _
	         + $m[6] * ( $m[1] * $m[5] - $m[4] * $m[2] )

	Local $r[9]

	$r[0] = + ( $m[4] * $m[8] - $m[7] * $m[5] ) / $d
	$r[3] = - ( $m[3] * $m[8] - $m[6] * $m[5] ) / $d
	$r[6] = + ( $m[3] * $m[7] - $m[6] * $m[4] ) / $d

	$r[1] = - ( $m[1] * $m[8] - $m[7] * $m[2] ) / $d
	$r[4] = + ( $m[0] * $m[8] - $m[6] * $m[2] ) / $d
	$r[7] = - ( $m[0] * $m[7] - $m[6] * $m[1] ) / $d

	$r[2] = + ( $m[1] * $m[5] - $m[4] * $m[2] ) / $d
	$r[5] = - ( $m[0] * $m[5] - $m[3] * $m[2] ) / $d
	$r[8] = + ( $m[0] * $m[4] - $m[3] * $m[1] ) / $d

	Return $r
EndFunc


; Multiplies two 3x3 matrices and returns a 3x3 matrix
Func glmMultiplyMat3( ByRef $m1, ByRef $m2 ) ; ( $mat3, $mat3 )
	; glm\core\type_mat3x3.inl line 591

	Local $m[16]

	$m[0] = $m1[0] * $m2[0] + $m1[3] * $m2[1] + $m1[6] * $m2[2]
	$m[1] = $m1[1] * $m2[0] + $m1[4] * $m2[1] + $m1[7] * $m2[2]
	$m[2] = $m1[2] * $m2[0] + $m1[5] * $m2[1] + $m1[8] * $m2[2]

	$m[3] = $m1[0] * $m2[3] + $m1[3] * $m2[4] + $m1[6] * $m2[5]
	$m[4] = $m1[1] * $m2[3] + $m1[4] * $m2[4] + $m1[7] * $m2[5]
	$m[5] = $m1[2] * $m2[3] + $m1[5] * $m2[4] + $m1[8] * $m2[5]

	$m[6] = $m1[0] * $m2[6] + $m1[3] * $m2[7] + $m1[6] * $m2[8]
	$m[7] = $m1[1] * $m2[6] + $m1[4] * $m2[7] + $m1[7] * $m2[8]
	$m[8] = $m1[2] * $m2[6] + $m1[5] * $m2[7] + $m1[8] * $m2[8]

	Return $m
EndFunc


; Returns the transposed matrix of a 3x3 matrix
Func glmTransposeMat3( ByRef $m ) ; ( $mat3 )
	; glm\core\func_matrix.hpp line 44
	; glm\core\func_matrix.inl line 227

	Local $r[9]

	$r[0] = $m[0]
	$r[1] = $m[3]
	$r[2] = $m[6]

	$r[3] = $m[1]
	$r[4] = $m[4]
	$r[5] = $m[7]

	$r[6] = $m[2]
	$r[7] = $m[5]
	$r[8] = $m[8]

	Return $r
EndFunc


; ------------------------------------- 2,3,4-component vectors --------------------------------------

;  $vec2[2] = [ $num1, $num2 ]
;  $vec3[3] = [ $num1, $num2, $num3 ]
;  $vec4[4] = [ $num1, $num2, $num3, $num4 ]


; ----------------------------------- 2-component vector functions -----------------------------------

Func glmLengthVec2Num( ByRef $vec2 )
	Return Sqrt( $vec2[0]*$vec2[0] + $vec2[1]*$vec2[1] )
EndFunc

Func glmNormalizeVec2( $vec2 )
	Local $l = glmLengthVec2Num( $vec2 )
	$vec2[0] /= $l
	$vec2[1] /= $l
	Return $vec2
EndFunc


; ----------------------------------- 3-component vector functions -----------------------------------

Func glmAddVec3( ByRef $x, ByRef $y ) ; ( $vec3, $vec3 )
	Local $vec3[3] = [ $x[0] + $y[0], $x[1] + $y[1], $x[2] + $y[2] ]
	Return $vec3
EndFunc

Func glmCrossVec3( ByRef $x, ByRef $y ) ; ( $vec3, $vec3 )
	; func_geometric.hpp line 48
	; func_geometric.inl line 179
	Local $vec3[3] = [ _
		$x[1] * $y[2] - $y[1] * $x[2], _
		$x[2] * $y[0] - $y[2] * $x[0], _
		$x[0] * $y[1] - $y[0] * $x[1] ]
	Return $vec3
EndFunc

Func glmDotVec3Num( ByRef $x, ByRef $y ) ; ( $vec3, $vec3 )
	; func_geometric.hpp line 39
	; func_geometric.inl line 139
	Return $x[0] * $y[0] + $x[1] * $y[1] + $x[2] * $y[2]
EndFunc

Func glmLengthVec3Num( ByRef $vec3 )
	Return Sqrt( $vec3[0]*$vec3[0] + $vec3[1]*$vec3[1] + $vec3[2]*$vec3[2] )
EndFunc

Func glmMultiplyMat3Vec( ByRef $m, ByRef $v ) ; ( $mat3, $vec3 )
	#cs
	Local $r[3] = [ $m[0] * $v[0] + $m[3] * $v[1] + $m[6] * $v[2], _
									$m[1] * $v[0] + $m[4] * $v[1] + $m[7] * $v[2], _
									$m[2] * $v[0] + $m[5] * $v[1] + $m[8] * $v[2] ]
	#ce
	Local $r[3] = [ $m[0] * $v[0] + $m[1] * $v[1] + $m[2] * $v[2], _
									$m[3] * $v[0] + $m[4] * $v[1] + $m[5] * $v[2], _
									$m[6] * $v[0] + $m[7] * $v[1] + $m[8] * $v[2] ]
	Return $r
EndFunc

Func glmNormalizeVec3( $vec3 )
	Local $l = glmLengthVec3Num( $vec3 )
	$vec3[0] /= $l
	$vec3[1] /= $l
	$vec3[2] /= $l
	Return $vec3
EndFunc

Func glmSubtractVec3( ByRef $x, ByRef $y ) ; ( $vec3, $vec3 )
	Local $vec3[3] = [ $x[0] - $y[0], $x[1] - $y[1], $x[2] - $y[2] ]
	Return $vec3
EndFunc


; ----------------------------------- 4-component vector functions -----------------------------------

Func glmAddVec4( ByRef $x, ByRef $y ) ; ( $vec4, $vec4 )
	Local $vec4[4] = [ $x[0] + $y[0], $x[1] + $y[1], $x[2] + $y[2], $x[3] + $y[3] ]
	Return $vec4
EndFunc

Func glmDotVec4Num( ByRef $x, ByRef $y ) ; ( $vec4, $vec4 )
	Return $x[0] * $y[0] + $x[1] * $y[1] + $x[2] * $y[2] + $x[3] * $y[3]
EndFunc

Func glmLengthVec4Num( ByRef $vec4 )
	Return Sqrt( $vec4[0]*$vec4[0] + $vec4[1]*$vec4[1] + $vec4[2]*$vec4[2] + $vec4[3]*$vec4[3] )
EndFunc

Func glmMixVec4( ByRef $x, ByRef $y, $a ) ; ( $vec4, $vec4, $num )
	Local $vec4[4] = [ _
		$x[0] + $a * ( $y[0] - $x[0] ), _
		$x[1] + $a * ( $y[1] - $x[1] ), _
		$x[2] + $a * ( $y[2] - $x[2] ), _
		$x[3] + $a * ( $y[3] - $x[3] ) ]
	Return $vec4
EndFunc

Func glmMultVec4( ByRef $x, $a ) ; ( $vec4, $num )
	Local $vec4[4] = [ $x[0] * $a, $x[1] * $a, $x[2] * $a, $x[3] * $a ]
	Return $vec4
EndFunc

Func glmNormalizeVec4( $vec4 )
	Local $l = glmLengthVec4Num( $vec4 )
	$vec4[0] /= $l
	$vec4[1] /= $l
	$vec4[2] /= $l
	$vec4[3] /= $l
	Return $vec4
EndFunc

Func glmSubtractVec4( ByRef $x, ByRef $y ) ; ( $vec4, $vec4 )
	Local $vec4[4] = [ $x[0] - $y[0], $x[1] - $y[1], $x[2] - $y[2], $x[3] - $y[3] ]
	Return $vec4
EndFunc


; ---------------------------------- 4-component quaternion vectors ----------------------------------

; Converts a 4-component quaternion vector into a 4x4 rotation matrix
Func glmCast( ByRef $q ) ; ( $vec4 )
	; glm\gtc\quaternion.hpp line 190
	; glm\gtc\quaternion.hpp line 503

	Local $m3 = glmCastMat3( $q )

	Local $m4[16] = [ $m3[0], $m3[1], $m3[2], 0, _
										$m3[3], $m3[4], $m3[5], 0, _
										$m3[6], $m3[7], $m3[8], 0, _
										   0,      0,      0,   1 ]
	Return $m4
EndFunc


; Converts a 4-component quaternion vector into a 3x3 matrix
Func glmCastMat3( ByRef $q ) ; ( $vec4 )
	; glm\gtc\quaternion.hpp line 184
	; glm\gtc\quaternion.hpp line 482

	Local $x = $q[0], $y = $q[1], $z = $q[2], $w = $q[3]
	Local $r[9]

	$r[0] = 1 - 2 * $y * $y - 2 * $z * $z
	$r[1] = 2 * $x * $y + 2 * $w * $z
	$r[2] = 2 * $x * $z - 2 * $w * $y

	$r[3] = 2 * $x * $y - 2 * $w * $z
	$r[4] = 1 - 2 * $x * $x - 2 * $z * $z
	$r[5] = 2 * $y * $z + 2 * $w * $x

	$r[6] = 2 * $x * $z + 2 * $w * $y
	$r[7] = 2 * $y * $z - 2 * $w * $x
	$r[8] = 1 - 2 * $x * $x - 2 * $y * $y

	Return $r
EndFunc


; Converts a 3x3 matrix to a 4-component quaternion vector
Func glmQuatCastMat3Vec4( ByRef $mat3 ) ; ( $mat3 )
	; glm\gtc\quaternion.hpp line 196
	; glm\gtc\quaternion.hpp line 512

	Local $fourXSquaredMinus1 = $mat3[0] - $mat3[4] - $mat3[8]
	Local $fourYSquaredMinus1 = $mat3[4] - $mat3[0] - $mat3[8]
	Local $fourZSquaredMinus1 = $mat3[8] - $mat3[0] - $mat3[4]
	Local $fourWSquaredMinus1 = $mat3[0] + $mat3[4] + $mat3[8]

	Local $biggestIndex = 0
	Local $fourBiggestSquaredMinus1 = $fourWSquaredMinus1
	If $fourXSquaredMinus1 > $fourBiggestSquaredMinus1 Then
		$fourBiggestSquaredMinus1 = $fourXSquaredMinus1
		$biggestIndex = 1
	EndIf
	If $fourYSquaredMinus1 > $fourBiggestSquaredMinus1 Then
		$fourBiggestSquaredMinus1 = $fourYSquaredMinus1
		$biggestIndex = 2
	EndIf
	If $fourZSquaredMinus1 > $fourBiggestSquaredMinus1 Then
		$fourBiggestSquaredMinus1 = $fourZSquaredMinus1
		$biggestIndex = 3
	EndIf

	Local $biggestVal = Sqrt( $fourBiggestSquaredMinus1 + 1 ) * 0.5
	Local $mult = 0.25 / $biggestVal

	Local $Result[4]
	Switch $biggestIndex
		case 0
			$Result[3] = $biggestVal
			$Result[0] = ( $mat3[5] - $mat3[7] ) * $mult
			$Result[1] = ( $mat3[6] - $mat3[2] ) * $mult
			$Result[2] = ( $mat3[1] - $mat3[3] ) * $mult
		case 1
			$Result[3] = ( $mat3[5] - $mat3[7] ) * $mult
			$Result[0] = $biggestVal
			$Result[1] = ( $mat3[1] + $mat3[3] ) * $mult
			$Result[2] = ( $mat3[6] + $mat3[2] ) * $mult
		case 2
			$Result[3] = ( $mat3[6] - $mat3[2] ) * $mult
			$Result[0] = ( $mat3[1] + $mat3[3] ) * $mult
			$Result[1] = $biggestVal
			$Result[2] = ( $mat3[5] + $mat3[7] ) * $mult
		case 3
			$Result[3] = ( $mat3[1] - $mat3[3] ) * $mult
			$Result[0] = ( $mat3[6] + $mat3[2] ) * $mult
			$Result[1] = ( $mat3[5] + $mat3[7] ) * $mult
			$Result[2] = $biggestVal
	EndSwitch

	Return $Result
EndFunc


; Converts a 4x4 matrix to a 4-component quaternion vector
Func glmQuatCastVec4( ByRef $mat4 ) ; ( $mat4 )
	Local $mat3[9] = [ $mat4[0], $mat4[1], $mat4[2], _
	                   $mat4[4], $mat4[5], $mat4[6], _
	                   $mat4[8], $mat4[9], $mat4[10] ]
	Return glmQuatCastMat3Vec4( $mat3 )
EndFunc


; Returns the conjugate (also a quaternion) of a quaternion vector
Func glmQuaternionConjugateVec4( ByRef $axis, $fAngle ) ; ( $vec3, $num )
	Local $v[4] = [ -$axis[0], -$axis[1], -$axis[2], $fAngle ]
	Return $v
EndFunc


; Returns a 4-component quaternion vector from an axis vector and an angle
Func glmQuaternionVec4( ByRef $axis, $fAngle ) ; ( $vec3, $num )
	Local $c = Cos( $fAngle * 0.5 ), $s = Sin( $fAngle * 0.5 )
	Local $v[4] = [ $axis[0] * $s, $axis[1] * $s, $axis[2] * $s, $c  ]
	Return $v
EndFunc


; Multiplies two 4-component quaternion vectors and returns a quaternion vector
Func glmQuaternionsMultVec4( ByRef $q, ByRef $p ) ; ( $vec4, $vec4 )
	Local $r[4] = [ _
		$q[3] * $p[0] + $q[0] * $p[3] + $q[1] * $p[2] - $q[2] * $p[1], _
		$q[3] * $p[1] + $q[1] * $p[3] + $q[2] * $p[0] - $q[0] * $p[2], _
		$q[3] * $p[2] + $q[2] * $p[3] + $q[0] * $p[1] - $q[1] * $p[0], _
		$q[3] * $p[3] - $q[0] * $p[0] - $q[1] * $p[1] - $q[2] * $p[2] ]
	Return $r
EndFunc


; ---------------------------------------- Utility functions -----------------------------------------

; Prints a 3-component vector in Scite console, 2 decimals
Func glmConsoleWriteVec3Dec2( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf1( $v[0] ) & sf1( $v[1] ) & sf2( $v[2] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4-component vector in Scite console, 2 decimals
Func glmConsoleWriteVec4Dec2( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2], [3] ) = " & sf1( $v[0] ) & sf1( $v[1] ) & sf1( $v[2] ) & sf2( $v[3] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 3x3 matrix in Scite console, 2 decimals
Func glmConsoleWriteMat3Dec2( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf1( $m[0] ) & sf1( $m[1] ) & sf2( $m[2] ) )
	ConsoleWrite( "( [3], [4], [5] ) = " & sf1( $m[3] ) & sf1( $m[4] ) & sf2( $m[5] ) )
	ConsoleWrite( "( [6], [7], [8] ) = " & sf1( $m[6] ) & sf1( $m[7] ) & sf2( $m[8] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4x4 matrix in Scite console, 2 decimals
Func glmConsoleWriteMat4Dec2( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "(  [0],  [1],  [2],  [3] ) = " & sf1(  $m[0] ) & sf1(  $m[1] ) & sf1(  $m[2] ) & sf2(  $m[3] ) )
	ConsoleWrite( "(  [4],  [5],  [6],  [7] ) = " & sf1(  $m[4] ) & sf1(  $m[5] ) & sf1(  $m[6] ) & sf2(  $m[7] ) )
	ConsoleWrite( "(  [8],  [9], [10], [11] ) = " & sf1(  $m[8] ) & sf1(  $m[9] ) & sf1( $m[10] ) & sf2( $m[11] ) )
	ConsoleWrite( "( [12], [13], [14], [15] ) = " & sf1( $m[12] ) & sf1( $m[13] ) & sf1( $m[14] ) & sf2( $m[15] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

Func sf1( $v )
	Return StringFormat( "%.2f", $v ) & ", "
EndFunc
Func sf2( $v )
	Return StringFormat( "%.2f", $v ) & @CRLF
EndFunc


; Prints a 3-component vector in Scite console, 4 decimals
Func glmConsoleWriteVec3Dec4( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf3( $v[0] ) & sf3( $v[1] ) & sf4( $v[2] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4-component vector in Scite console, 4 decimals
Func glmConsoleWriteVec4Dec4( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2], [3] ) = " & sf3( $v[0] ) & sf3( $v[1] ) & sf3( $v[2] ) & sf4( $v[3] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 3x3 matrix in Scite console, 4 decimals
Func glmConsoleWriteMat3Dec4( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf3( $m[0] ) & sf3( $m[1] ) & sf4( $m[2] ) )
	ConsoleWrite( "( [3], [4], [5] ) = " & sf3( $m[3] ) & sf3( $m[4] ) & sf4( $m[5] ) )
	ConsoleWrite( "( [6], [7], [8] ) = " & sf3( $m[6] ) & sf3( $m[7] ) & sf4( $m[8] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4x4 matrix in Scite console, 4 decimals
Func glmConsoleWriteMat4Dec4( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "(  [0],  [1],  [2],  [3] ) = " & sf3(  $m[0] ) & sf3(  $m[1] ) & sf3(  $m[2] ) & sf4(  $m[3] ) )
	ConsoleWrite( "(  [4],  [5],  [6],  [7] ) = " & sf3(  $m[4] ) & sf3(  $m[5] ) & sf3(  $m[6] ) & sf4(  $m[7] ) )
	ConsoleWrite( "(  [8],  [9], [10], [11] ) = " & sf3(  $m[8] ) & sf3(  $m[9] ) & sf3( $m[10] ) & sf4( $m[11] ) )
	ConsoleWrite( "( [12], [13], [14], [15] ) = " & sf3( $m[12] ) & sf3( $m[13] ) & sf3( $m[14] ) & sf4( $m[15] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

Func sf3( $v )
	Return StringFormat( "%.4f", $v ) & ", "
EndFunc
Func sf4( $v )
	Return StringFormat( "%.4f", $v ) & @CRLF
EndFunc


; Prints a 3-component vector in Scite console, 8 decimals
Func glmConsoleWriteVec3Dec8( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf5( $v[0] ) & sf5( $v[1] ) & sf6( $v[2] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4-component vector in Scite console, 8 decimals
Func glmConsoleWriteVec4Dec8( ByRef $v, $header = "", $bLF = 0 ) ; ( $vec4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2], [3] ) = " & sf5( $v[0] ) & sf5( $v[1] ) & sf5( $v[2] ) & sf6( $v[3] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 3x3 matrix in Scite console, 8 decimals
Func glmConsoleWriteMat3Dec8( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat3, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "( [0], [1], [2] ) = " & sf5( $m[0] ) & sf5( $m[1] ) & sf6( $m[2] ) )
	ConsoleWrite( "( [3], [4], [5] ) = " & sf5( $m[3] ) & sf5( $m[4] ) & sf6( $m[5] ) )
	ConsoleWrite( "( [6], [7], [8] ) = " & sf5( $m[6] ) & sf5( $m[7] ) & sf6( $m[8] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

; Prints a 4x4 matrix in Scite console, 8 decimals
Func glmConsoleWriteMat4Dec8( ByRef $m, $header = "", $bLF = 0 ) ; ( $mat4, $header, 0/1 )
	If $header <> "" Then ConsoleWrite( $header & @CRLF )
	ConsoleWrite( "(  [0],  [1],  [2],  [3] ) = " & sf5(  $m[0] ) & sf5(  $m[1] ) & sf5(  $m[2] ) & sf6(  $m[3] ) )
	ConsoleWrite( "(  [4],  [5],  [6],  [7] ) = " & sf5(  $m[4] ) & sf5(  $m[5] ) & sf5(  $m[6] ) & sf6(  $m[7] ) )
	ConsoleWrite( "(  [8],  [9], [10], [11] ) = " & sf5(  $m[8] ) & sf5(  $m[9] ) & sf5( $m[10] ) & sf6( $m[11] ) )
	ConsoleWrite( "( [12], [13], [14], [15] ) = " & sf5( $m[12] ) & sf5( $m[13] ) & sf5( $m[14] ) & sf6( $m[15] ) )
	If $bLF Then ConsoleWrite( @CRLF )
EndFunc

Func sf5( $v )
	Return StringFormat( "%.8f", $v ) & ", "
EndFunc
Func sf6( $v )
	Return StringFormat( "%.8f", $v ) & @CRLF
EndFunc
