; Functions:
;   EnableOpenGL()        - Create a simple OpenGL rendering context for the highest supported version
;   EnableOpenGLEx()      - Create a proper OpenGL rendering context for the highest supported version
;                           Create a proper OpenGL version 3.3, 3.2, 3.1, 3.0 or 2.1 rendering context
;                           Create a generic OpenGL version 1.1 rendering context
;   CreateOpenGLContext() - A wrapper around EnableOpenGLEx() which falls back to EnableOpenGL() in case of an error
;   OpenGLProperContext() - A wrapper around EnableOpenGLEx() which prints out error messages
;   OpenGLSimpleContext() - A wrapper around EnableOpenGL() which prints out error messages
;   DisableOpenGL         - Disable an OpenGL rendering context

; http://www.opengl.org/wiki/Creating_an_OpenGL_Context_(WGL)
; http://www.opengl.org/registry/specs/ARB/wgl_pixel_format.txt
; http://www.opengl.org/registry/specs/ARB/wgl_create_context.txt


#include-once

#include "oglUtilities.au3"


; $aContextCreationWarning[2][2]
; If wglCreateContextAttribsARB() can't be used or fails to create a context in EnableOpenGLEx() the usual function wglCreateContext()
; will be used instead. In this case, the error which caused the function to fail, is regarded as a warning and will be stored in the
; array. This is neccessary to be able to print out the warning in the wrapper functions. If you use EnableOpenGLEx() directly without
; a wrapper function you have to inspect the array yourself.
;
; $aContextCreationWarning[0][0] : 0 => no warnings, 1 => one warning
; $aContextCreationWarning[1][0] = @error
; $aContextCreationWarning[1][1] = @extended
; There are 4 different warnings. See Warning 15 - 18 below.
; But there can only be one at a time.
;
; (There is only used one number sequence for both errors and warnings: 1 - 14 are errors, 15 - 18 are warnings, 19 - 21 are errors.)

Global $aContextCreationWarning[2][2]


Global Const $tagPIXELFORMATDESCRIPTOR = _
	"word  Size;" & _
	"word  Version;" & _
	"dword Flags;" & _
	"byte  PixelType;" & _
	"byte  ColorBits;" & _
	"byte  RedBits;" & _
	"byte  RedShift;" & _
	"byte  GreenBits;" & _
	"byte  GreenShift;" & _
	"byte  BlueBits;" & _
	"byte  BlueShift;" & _
	"byte  AlphaBits;" & _
	"byte  AlphaShift;" & _
	"byte  AccumBits;" & _
	"byte  AccumRedBits;" & _
	"byte  AccumGreenBits;" & _
	"byte  AccumBlueBits;" & _
	"byte  AccumAlphaBits;" & _
	"byte  DepthBits;" & _
	"byte  StencilBits;" & _
	"byte  AuxBuffers;" & _
	"byte  LayerType;" & _
	"byte  Reserved;" & _
	"dword LayerMask;" & _
	"dword VisibleMask;" & _
	"dword DamageMask"


; Create a simple (simple is the term used in the OpenGL wiki) OpenGL rendering context
Func EnableOpenGL( $hWnd, ByRef $hDeviceContext, ByRef $hOpenGLRenderingContext )

	Local Const $PFD_TYPE_RGBA = 0
	Local Const $PFD_MAIN_PLANE = 0
	Local Const $PFD_DOUBLEBUFFER = 1
	Local Const $PFD_DRAW_TO_WINDOW = 4
	Local Const $PFD_SUPPORT_OPENGL = 32

	Local $tPIXELFORMATDESCRIPTOR = DllStructCreate( $tagPIXELFORMATDESCRIPTOR )

	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "Size", DllStructGetSize( $tPIXELFORMATDESCRIPTOR ) )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "Version", 1 )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "Flags", BitOR( $PFD_DRAW_TO_WINDOW, $PFD_SUPPORT_OPENGL, $PFD_DOUBLEBUFFER ) )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "PixelType", $PFD_TYPE_RGBA )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "ColorBits", 32 )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "DepthBits", 24 )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "StencilBits", 8 )
	DllStructSetData( $tPIXELFORMATDESCRIPTOR, "LayerType", $PFD_MAIN_PLANE )

	$hDeviceContext = GetDC( $hWnd )
	If @error Then Return SetError( 1, 0, 0 )

	Local $iFormat = ChoosePixelFormat( $hDeviceContext, DllStructGetPtr( $tPIXELFORMATDESCRIPTOR ) )
	If @error Then Return SetError( 2, @extended, 0 )

	SetPixelFormat( $hDeviceContext, $iFormat, DllStructGetPtr( $tPIXELFORMATDESCRIPTOR ) )
	If @error Then Return SetError( 3, @extended, 0 )

	$hOpenGLRenderingContext = wglCreateContext( $hDeviceContext )
	If @error Then Return SetError( 4, @extended, 0 )

	wglMakeCurrent( $hDeviceContext, $hOpenGLRenderingContext )
	If @error Then Return SetError( 5, @extended, 0 )

	Return SetError( 0, 0, 1 )

EndFunc


; Create a proper (proper is the term used in the OpenGL wiki) OpenGL rendering context
; You can specify a major and minor version to be used with wglCreateContextAttribsARB()
; ( major, minor ) = ( -1, -1 ) : Create a proper context for the highest supported version
; ( 2, 1 ), ( 3, 0 ) - ( 3, 3 ) : Create a proper context for the specific version
; ( major, minor ) = ( 1, 1 )   : Create a generic version 1.1 context
; Other major and minor values will generate an error
Func EnableOpenGLEx( $hWnd, ByRef $hDeviceContext, ByRef $hOpenGLRenderingContext, $iMajorVersion = -1, $iMinorVersion = -1, $sContext = "Default" )

	; Reset warnings
	$aContextCreationWarning[0][0] = 0

	; Check $iMajorVersion and $iMinorVersion
	Local $fVersion = True
	Select
		Case $iMajorVersion = -1 And $iMinorVersion = -1 ; This indicates highest supported version, only way to get a 4.0+ proper context
		Case $iMajorVersion =  1 And $iMinorVersion =  1
		Case $iMajorVersion =  2 And $iMinorVersion =  1
		Case $iMajorVersion =  3 And ( $iMinorVersion >= 0 And $iMinorVersion <= 3 )
		Case Else
			$fVersion = False
	EndSelect
	If Not $fVersion Then Return SetError( 6, 0, 0 )

	; Create fake context
	Local $hGUIfake, $hDCfake, $hRCfake
	$hGUIfake = GUICreate( "Fake context" )
	EnableOpenGL( $hGUIfake, $hDCfake, $hRCfake ) ; Error codes 1 - 5 are used here
	If @error Then Return SetError( @error, @extended, 0 )

	; If the major version number for the fake context is less than 3 we can't create a proper context
	Local $sVersion = StringLeft( glGetString( $GL_VERSION ), 3 )
	Local $iFakeMajor = Number( StringLeft( $sVersion, 1 ) )
	Local $iFakeMinor = Number( StringRight( $sVersion, 1 ) )
	If $iFakeMajor < 3 Then Return SetError( 7, $iFakeMajor, 0 )

	; $iFakeMajor is the major version number for the highest supported SIMPLE version
	; This will match the major version number for the highest supported PROPER version
	; Similarly for the minor version number

	; If $iMajorVersion = 3 and $iFakeMajor = 3 and $iFakeMinor < $iMinorVersion we can't create a proper context
	If $iMajorVersion = 3 And $iFakeMajor = 3 And $iFakeMinor < $iMinorVersion Then Return SetError( 8, $iFakeMinor, 0 )

	; Device context
	$hDeviceContext = GetDC( $hWnd )
	If @error Then Return SetError( 9, 0, 0 )

	; To create a proper context we must be able to choose a proper pixel format with wglChoosePixelFormatARB().
	; This function belongs to the WGL_ARB_pixel_format extension. Verify that this extension is available.
	If Not CheckWGLextension( "WGL_ARB_pixel_format", $hDeviceContext ) Then _
		Return SetError( 11, 0, 0 ) ; The WGL_ARB_pixel_format extension isn't available

	; Pixel format attribute list. For available attributes see
	; http://www.opengl.org/registry/specs/ARB/wgl_pixel_format.txt.

	Local $aPixelFormatAttribList

	Switch $sContext
		Case "MSAAbest"
			If Not CheckWGLextension( "WGL_ARB_multisample" ) _
				 And Not CheckOpenGLextension( "GL_ARB_multisample" ) Then _
					 Return SetError( 11, 0, 0 ) ; The WGL_ARB_multisample extension isn't available
			$aPixelFormatAttribList = _
				GetPixelFormatAttribListMSAAbest( $hDeviceContext )
			If Not IsArray( $aPixelFormatAttribList ) Then _
				Return SetError( 13, 0, 0 ) ; wglChoosePixelFormatARB() could not select a pixel format

		Case "MSAA"
			If Not CheckWGLextension( "WGL_ARB_multisample" ) _
				 And Not CheckOpenGLextension( "GL_ARB_multisample" ) Then _
					 Return SetError( 11, 0, 0 ) ; The WGL_ARB_multisample extension isn't available
			Local $aPixelFormatAttribListMSAA[19] = [ _
        $WGL_DRAW_TO_WINDOW_ARB, $GL_TRUE, _
        $WGL_ACCELERATION_ARB,   $WGL_FULL_ACCELERATION_ARB, _
        $WGL_COLOR_BITS_ARB,     24, _
        $WGL_ALPHA_BITS_ARB,     8, _
        $WGL_DEPTH_BITS_ARB,     24, _
        $WGL_STENCIL_BITS_ARB,   8, _
        $WGL_DOUBLE_BUFFER_ARB,  $GL_TRUE, _
        $WGL_SAMPLE_BUFFERS_ARB, 1, _
        $WGL_SAMPLES_ARB,        4, _
				0 ] ; End of attributes list
			$aPixelFormatAttribList = _
				$aPixelFormatAttribListMSAA

		Case "Default"
			Local $aPixelFormatAttribListDefault[15] = [ _
				$WGL_DRAW_TO_WINDOW_ARB, $GL_TRUE, _
				$WGL_SUPPORT_OPENGL_ARB, $GL_TRUE, _
				$WGL_DOUBLE_BUFFER_ARB,  $GL_TRUE, _
				$WGL_PIXEL_TYPE_ARB,     $WGL_TYPE_RGBA_ARB, _
				$WGL_COLOR_BITS_ARB,     32, _
				$WGL_DEPTH_BITS_ARB,     24, _  ; Set the depth buffer to 32 and you will get this context: Version:  1.1.0
				$WGL_STENCIL_BITS_ARB,   8, _   ;                                                           Vendor:   Microsoft Corporation
				0 ] ; End of attributes list    ;                                                           Renderer: GDI Generic
			If $iMajorVersion = 1 Then _
				$aPixelFormatAttribListDefault[11] = 32
			$aPixelFormatAttribList = _
				$aPixelFormatAttribListDefault
	EndSwitch
	
	Local $n = UBound( $aPixelFormatAttribList )
	Local $tPixelFormatAttribList = DllStructCreate( "int[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tPixelFormatAttribList, 1, $aPixelFormatAttribList[$i], $i+1 )
	Next

	; Choose pixel format
	Local $iPixelFormat, $iNumFormats
	wglChoosePixelFormatARB( $hDeviceContext, DllStructGetPtr( $tPixelFormatAttribList ), 0, 1, $iPixelFormat, $iNumFormats )
	If @error Then Return SetError( 11+@error, @extended, 0 )
	; 11+@error = 12 if wglGetProcAddress fails
	; 11+@error = 13 if wglChoosePixelFormatARB fails

	; Set pixel format
	Local $tPIXELFORMATDESCRIPTOR = DllStructCreate( $tagPIXELFORMATDESCRIPTOR )
	SetPixelFormat( $hDeviceContext, $iPixelFormat, DllStructGetPtr( $tPIXELFORMATDESCRIPTOR ) )
	If @error Then Return SetError( 14, @extended, 0 )

	; At this point we have a proper pixel format. Next step is to create a proper context.

	; For a proper context it's recommended to use wglCreateContextAttribsARB() in the WGL_ARB_create_context
	; extension.

	; For 3.2+ versions to be backward compatible it's neccessary to set the WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
	; bit for WGL_CONTEXT_PROFILE_MASK_ARB in the context attribute list. This needs the WGL_ARB_create_context_profile
	; extension.

	; If the conditions for using wglCreateContextAttribsARB() is not met or the function fails, the usual wglCreateContext()
	; will be used instead.

	Local $fConditions = False

	Local $f32PlusVersions = False
	If ( $iMajorVersion = 3 And $iMinorVersion >= 2 ) Or _
		 ( $iMajorVersion = -1 And $iFakeMajor = 3 And $iFakeMinor >= 2 ) Or _
		 ( $iMajorVersion = -1 And $iFakeMajor > 3 ) Then _ ; 4.0+ versions
		 $f32PlusVersions = True

	; Check for WGL_ARB_create_context
	If CheckWGLextension( "WGL_ARB_create_context" ) Then _
		$fConditions = True
	If Not $fConditions Then _
		SetContextCreationWarning( 15, 0 ) ; Warning 15
		; The WGL_ARB_create_context extension isn't available
		; Generate a warning and fall back to wglCreateContext()

	If $fConditions And $f32PlusVersions Then
		; Check for WGL_ARB_create_context_profile
		$fConditions = False
		If CheckWGLextension( "WGL_ARB_create_context_profile" ) Then _
			$fConditions = True
		If Not $fConditions Then _
			SetContextCreationWarning( 16, 0 ) ; Warning 16
			; The WGL_ARB_create_context_profile extension isn't available
			; Generate a warning and fall back to wglCreateContext()
	EndIf

	If $fConditions Then

		; Use wglCreateContextAttribsARB()

		; Context attribute list. For available attributes see
		; http://www.opengl.org/registry/specs/ARB/wgl_create_context.txt

		; Set major and minor version numbers if $iMajorVersion = -1 ( and $iMinorVersion = -1 ).
		; $iMajorVersion = -1 means get a PROPER context for the highest supported version. The
		; highest supported PROPER version matches the highest supported SIMPLE version.
		; I.e. the version for the fake context.
		If $iMajorVersion = -1 Then				; The only way to get a 4.0+ version is to use $iMajorVersion = -1
			$iMajorVersion = $iFakeMajor		; A 4.0+ version will always be the highest supported version
			$iMinorVersion = $iFakeMinor
		EndIf

		Local $aContextAttribs[7] = [ _
			$WGL_CONTEXT_MAJOR_VERSION_ARB, $iMajorVersion, _
			$WGL_CONTEXT_MINOR_VERSION_ARB, $iMinorVersion, _
			$WGL_CONTEXT_FLAGS_ARB, 0, _
			0 ] ; End of attributes list

		If $f32PlusVersions Then
			ReDim $aContextAttribs[9]
			$aContextAttribs[6] = $WGL_CONTEXT_PROFILE_MASK_ARB
			$aContextAttribs[7] = $WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
			$aContextAttribs[8] = 0
		EndIf

		$n = UBound( $aContextAttribs )
		Local $tContextAttribs = DllStructCreate( "int[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $tContextAttribs, 1, $aContextAttribs[$i], $i+1 )
		Next

		$hOpenGLRenderingContext = wglCreateContextAttribsARB( $hDeviceContext, 0, DllStructGetPtr( $tContextAttribs ) )
		If @error Then
			SetContextCreationWarning( 16+@error, @extended )
			; 16+@error = 17 if wglGetProcAddress fails						; Warning 17
			; 16+@error = 18 if wglCreateContextAttribsARB fails	; Warning 18

			; Fall back to wglCreateContext()
			$hOpenGLRenderingContext = wglCreateContext( $hDeviceContext )
			If @error Then Return SetError( 19, @extended, 0 )
		EndIf

	Else

		; Use wglCreateContext()

		$hOpenGLRenderingContext = wglCreateContext( $hDeviceContext )
		If @error Then Return SetError( 20, @extended, 0 )

	EndIf

	; Make current context
	wglMakeCurrent( $hDeviceContext, $hOpenGLRenderingContext )
	If @error Then Return SetError( 21, @extended, 0 )

	; Delete fake context
	wglDeleteContext( $hRCfake )
	ReleaseDC( $hGUIfake, $hDCfake )
	GUIDelete( $hGUIfake )

	Return SetError( 0, 0, 1 )

EndFunc


Func GetPixelFormatAttribListMSAAbest( $hDeviceContext )
	Local $aPixelFormatAttribListMSAAbest[19] = [ _
		$WGL_DRAW_TO_WINDOW_ARB, $GL_TRUE, _
		$WGL_ACCELERATION_ARB,   $WGL_FULL_ACCELERATION_ARB, _
		$WGL_COLOR_BITS_ARB,     24, _
		$WGL_ALPHA_BITS_ARB,     8, _
		$WGL_DEPTH_BITS_ARB,     24, _
		$WGL_STENCIL_BITS_ARB,   8, _
		$WGL_DOUBLE_BUFFER_ARB,  $GL_TRUE, _
		$WGL_SAMPLE_BUFFERS_ARB, 1, _
		$WGL_SAMPLES_ARB,        0, _
		0 ] ; End of attributes list

	Local $n = UBound( $aPixelFormatAttribListMSAAbest )
	Local $tPixelFormatAttribList = DllStructCreate( "int[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tPixelFormatAttribList, 1, $aPixelFormatAttribListMSAAbest[$i], $i+1 )
	Next

	; Choose best MSAA pixel format
	Local $iPixelFormat, $iNumFormats, $iSamples = 16
	While $iSamples >= 1
		$aPixelFormatAttribListMSAAbest[17] = $iSamples
		DllStructSetData( $tPixelFormatAttribList, 1, $aPixelFormatAttribListMSAAbest[17], 18 )
		wglChoosePixelFormatARB( $hDeviceContext, DllStructGetPtr( $tPixelFormatAttribList ), 0, 1, $iPixelFormat, $iNumFormats )
		If Not @error And $iNumFormats Then ExitLoop
		$iSamples /= 2
	WEnd

	If $iSamples >= 1 Then
		;MsgBox( 0, "MSAA context: $iSamples", $iSamples )
		Return $aPixelFormatAttribListMSAAbest
	Else
		Return 0
	EndIf
EndFunc


Func DisableOpenGL( $hWnd, $hDC, $hglrc )
	wglMakeCurrent( 0, 0 )
	wglDeleteContext( $hglrc )
	ReleaseDC( $hWnd, $hDC )
EndFunc


; A wrapper around EnableOpenGL() which prints out error messages
Func OpenGLSimpleContext( $hWnd, ByRef $hDeviceContext, ByRef $hOpenGLRenderingContext )
	If Not EnableOpenGL( $hWnd, $hDeviceContext, $hOpenGLRenderingContext ) Then
		Local $aError = GetContextCreationError( 1, @error, @extended )
		MsgBox( 16, $aError[0], $aError[1] ) ; ScriptName, Error
		ConsoleWrite( $aError[1] & @CRLF )
		Return 0
	Else
		Return 1
	EndIf
EndFunc


; A wrapper around EnableOpenGLEx() which prints out error messages
Func OpenGLProperContext( $hWnd, ByRef $hDeviceContext, ByRef $hOpenGLRenderingContext, $iMajorVersion = -1, $iMinorVersion = -1 )
	If Not EnableOpenGLEx( $hWnd, $hDeviceContext, $hOpenGLRenderingContext, $iMajorVersion, $iMinorVersion ) Then
		Local $iError = @error, $iExtended = @extended
		If $aContextCreationWarning[0][0] Then GetContextCreationWarning()
		Local $aError = GetContextCreationError( 2, $iError, $iExtended )
		MsgBox( 16, $aError[0], $aError[1] ) ; ScriptName, Error
		ConsoleWrite( $aError[1] & @CRLF )
		Return 0
	Else
		If $aContextCreationWarning[0][0] Then _
			GetContextCreationWarning()
		Return 1
	EndIf
EndFunc


; A wrapper around EnableOpenGLEx() which falls back to EnableOpenGL() in case of an error
Func CreateOpenGLContext( $hWnd, ByRef $hDeviceContext, ByRef $hOpenGLRenderingContext, $iMajorVersion = -1, $iMinorVersion = -1 )

	If Not EnableOpenGLEx( $hWnd, $hDeviceContext, $hOpenGLRenderingContext, $iMajorVersion, $iMinorVersion ) Then

		; EnableOpenGLEx (proper context) failed
		Local $aError1 = GetContextCreationError( 2, @error, @extended )

		; Warnings to MsgBox and Scite console
		If $aContextCreationWarning[0][0] Then GetContextCreationWarning()

		; Fall back to EnableOpenGL (simple context)
		If Not EnableOpenGL( $hWnd, $hDeviceContext, $hOpenGLRenderingContext ) Then

			; EnableOpenGL failed too
			; Errors to MsgBox and Scite console
			Local $aError2 = GetContextCreationError( 1, @error, @extended )
			MsgBox( 16, $aError1[0], $aError1[1] & @CRLF & @CRLF & $aError2[1] )
			ConsoleWrite( $aError1[1] & @CRLF & @CRLF & $aError2[1] & @CRLF )
			Return 0

		Else

			; EnableOpenGL succeeded
			; EnableOpenGLEx errors to Scite console
			ConsoleWrite( $aError1[1] & @CRLF )
			ConsoleWrite( "Falls back to a simple context creation." & @CRLF )
			ConsoleWrite( "Function EnableOpenGL() succeeded." & @CRLF )
			Return 2

		EndIf

	Else

		; EnableOpenGLEx succeeded
		If $aContextCreationWarning[0][0] Then _
			GetContextCreationWarning()
		Return 1

	EndIf

EndFunc


Func SetContextCreationWarning( $iWarn, $iExtended )
	$aContextCreationWarning[0][0] = 1
	$aContextCreationWarning[1][0] = $iWarn
	$aContextCreationWarning[1][1] = $iExtended
EndFunc


Func GetContextCreationWarning()
	Local $aWarn, $iWarn, $iExtended
	$iWarn = $aContextCreationWarning[1][0]
	$iExtended = $aContextCreationWarning[1][1]
	Local $aWarn = GetContextCreationError( 2, $iWarn, $iExtended )
	MsgBox( 48, $aWarn[0], $aWarn[1] ) ; ScriptName, Warning
	ConsoleWrite( $aWarn[1] & @CRLF )
EndFunc


Func GetContextCreationError( $fFunc, $iError, $iExtended )

	Local Static $aErrorMessages[21] = [ _
		"GetDC() failed in creating a device context.", _																									;  1  ; EnableOpenGL(): 1 - 5
		"ChoosePixelFormat() could not select a pixel format.", _																					;  2
		"SetPixelFormat() could not set pixel format.", _																									;  3
		"wglCreateContext() failed in creating a rendering context.", _																		;  4
		"wglMakeCurrent() could not make rendering context current.", _																		;  5
		"Illegal major or minor version. Legal values are (1,1), (2,1), (3,0) - (3,3) or none.", _				;  6  ; EnableOpenGLEx(): 6 - 17
		"Major version number for fake context must be 3 or higher.", _																		;  7
		"Minor number for proper context must not be higher than minor number for fake context.", _				;  8
		"GetDC() failed in creating a device context.", _																									;  9
		"Could not get list of WGL extensions.", _																												; 10
		"Could not locate WGL_ARB_pixel_format or WGL_ARB_multisample in WGL extensions list.", _					; 11
		"wglGetProcAddress() could not locate address for wglChoosePixelFormatARB().", _									; 12
		"wglChoosePixelFormatARB() could not select a pixel format.", _																		; 13
		"SetPixelFormat() could not set pixel format.", _																									; 14
		"Could not locate WGL_ARB_create_context in WGL extensions list.", _															; 15  ; Warning
		"Could not locate WGL_ARB_create_context_profile in WGL extensions list.", _											; 16  ; Warning
		"wglGetProcAddress() could not locate address for wglCreateContextAttribsARB().", _								; 17  ; Warning
		"wglCreateContextAttribsARB() failed in creating a rendering context.", _													; 18  ; Warning
		"wglCreateContextAttribsARB() and wglCreateContext() failed in creating a rendering context.", _	; 19
		"wglCreateContext() failed in creating a rendering context.", _																		; 20
		"wglMakeCurrent() could not make rendering context current." ]																		; 21

	Local $sErrStr
	If $fFunc = 1 Then
		$sErrStr = "Function EnableOpenGL()" & @CRLF
	ElseIf $fFunc = 2 Then
		$sErrStr = "Function EnableOpenGLEx()" & @CRLF
		If $iError < 6 Then _
			$sErrStr &= "Create fake context with EnableOpenGL()" & @CRLF
	EndIf

	$sErrStr &= $aErrorMessages[$iError-1] & @CRLF

	Switch $iError

		Case 7 ; Error
			If $iExtended <> 0 Then _
				$sErrStr &= " Fake context major version number: " & $iExtended & "."
			$sErrStr &= "Error code: " & $iError & "."

		Case 8 ; Error
			If $iExtended <> 0 Then _
				$sErrStr &= " Fake context minor version number: " & $iExtended & "."
			$sErrStr &= "Error code: " & $iError & "."

		Case 15 To 18 ; Warnings
			$sErrStr &= "Can't use wglCreateContextAttribsARB(). Trying with wglCreateContext()." & @CRLF
			$sErrStr &= "Warning code: " & $iError & "."
			If $iExtended <> 0 Then _
				$sErrStr &= " Extended warning code (GetLastError): " & $iExtended & "."

		Case Else ; Errors
			$sErrStr &= "Error code: " & $iError & "."
			If $iExtended <> 0 Then _
				$sErrStr &= " Extended error code (GetLastError): " & $iExtended & "."

	EndSwitch

	Local $sScriptName = StringLeft( @ScriptName, StringInStr( @ScriptName, ".", 0, -1 ) - 1 )

	Local $aError[2] = [ $sScriptName, $sErrStr ]
	Return $aError

EndFunc
