; OpenGl functions from gl.h, glu.h and glew.h

; --- References ---

; OpenGL 1.1: http://msdn.microsoft.com/en-us/library/windows/desktop/dd374278(v=vs.85).aspx
; OpenGL 2.1: http://www.opengl.org/sdk/docs/man2/
; OpenGL 3.3: http://www.opengl.org/sdk/docs/man3/

; --- Error checking and error codes ---

; According to references most functions where the name ends on a type specification and a "v" ( <function>{bv,dv,fv,iv,sv,ubv,uiv,usv} )
; accepts a pointer to an array as a parameter. In translation to AutoIt these functions can accept both a pointer and an array. See
; glColor function below as an example. If the parameter isn't a pointer or an array an error value of 1 is returned ( @error = 1 ).
; This is implemented when the parameter is an input parameter.

; For functions that accept an array as a parameter the array is checked for size and dimension. In case of an error an error value
; of 1 is returned ( @error = 1 ).

; If wglGetProcAddress fails an error value of 2 is returned ( @error = 2 ) and the extended error value ( @extended ) is set to the
; error code returned by GetLastError.

; If a DllCall fails an error value of 3 is returned ( @error = 3 ).


#include-once

#include "..\Includes\Fasm\Utilities.au3"

; ------------------------------ OpenGL 1.1 ------------------------------

Func glAlphaFunc( $func, $ref )
	DllCall( $dllOpenGL32, "none", "glAlphaFunc", "uint", $func, "float", $ref )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glBegin( $mode )
	DllCall( $dllOpenGL32, "none", "glBegin", "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glBindTexture( $target, $texture )
	DllCall( $dllOpenGL32, "none", "glBindTexture", "uint", $target, "uint", $texture )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glBlendFunc( $sfactor, $dfactor )
	DllCall( $dllOpenGL32, "none", "glBlendFunc", "uint", $sfactor, "uint", $dfactor )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glCallList( $list )
	DllCall( $dllOpenGL32, "none", "glCallList", "uint", $list )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glCallLists( $n, $type, $lists )
	If IsString( $lists ) Then
		If $type <> $GL_UNSIGNED_BYTE Then Return SetError( 1, 0, 0 )
		DllCall( $dllOpenGL32, "none", "glCallLists", "int", $n, "uint", $type, "str", $lists )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $lists ) And UBound( $lists, 0 ) = 1 Then
		Local $sType, $l = UBound( $lists ), $struct
		Switch $type
			Case $GL_BYTE, $GL_UNSIGNED_BYTE
				$sType = "byte"
			Case $GL_SHORT
				$sType = "short"
			Case $GL_UNSIGNED_SHORT
				$sType = "ushort"
			Case $GL_INT
				$sType = "int"
			Case $GL_UNSIGNED_INT
				$sType = "uint"
			Case $GL_FLOAT
				$sType = "float"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		$struct = DllStructCreate( $sType & "[" & $l & "]" )
		For $i = 0 To $l - 1
			DllStructSetData( $struct, 1, $lists[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glCallLists", "int", $n, "uint", $type, "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glClear( $mask )
	DllCall( $dllOpenGL32, "none", "glClear", "uint", $mask )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glClearColor( $red, $green, $blue, $alpha )
	DllCall( $dllOpenGL32, "none", "glClearColor", "float", $red, "float", $green, "float", $blue, "float", $alpha )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glClearDepth( $depth )
	DllCall( $dllOpenGL32, "none", "glClearDepth", "double", $depth )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glColor3f( $red, $green, $blue )
	DllCall( $dllOpenGL32, "none", "glColor3f", "float", $red, "float", $green, "float", $blue )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glColor3fv( ByRef $v )
	If IsPtr( $v ) Then
		DllCall( $dllOpenGL32, "none", "glColor3fv", "ptr", $v )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $v ) And UBound( $v ) = 3 And UBound( $v, 0 ) = 1 Then
		Local $stColor = DllStructCreate( "float[3]" )
		For $i = 0 To 2
			DllStructSetData( $stColor, 1, $v[$i], $i+1 )
		Next
		DllCall( $dllOpenGL32, "none", "glColor3fv", "struct*", $stColor )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glColor3ub( $red, $green, $blue )
	DllCall( $dllOpenGL32, "none", "glColor3ub", "ubyte", $red, "ubyte", $green, "ubyte", $blue )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glColor4f( $red, $green, $blue, $alpha )
	DllCall( $dllOpenGL32, "none", "glColor4f", "float", $red, "float", $green, "float", $blue, "float", $alpha )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glColorPointer( $size, $type, $stride, $pointer )
	If IsPtr( $pointer ) Then
		DllCall( $dllOpenGL32, "none", "glColorPointer", "int", $size, "uint", $type, "int", $stride, "ptr", $pointer )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $pointer ) And UBound( $pointer, 0 ) = 1 Then
		Local $sType, $n = UBound( $pointer )
		Local Static $struct
		Switch $type
			Case $GL_BYTE, $GL_UNSIGNED_BYTE
				$sType = "byte"
			Case $GL_SHORT
				$sType = "short"
			Case $GL_UNSIGNED_SHORT
				$sType = "ushort"
			Case $GL_INT
				$sType = "int"
			Case $GL_UNSIGNED_INT
				$sType = "uint"
			Case $GL_FLOAT
				$sType = "float"
			Case $GL_DOUBLE
				$sType = "double"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		$struct = DllStructCreate( $sType & "[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $struct, 1, $pointer[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glColorPointer", "int", $size, "uint", $type, "int", $stride, "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glCullFace( $mode )
	DllCall( $dllOpenGL32, "none", "glCullFace", "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDeleteLists( $list, $range )
	DllCall( $dllOpenGL32, "none", "glDeleteLists", "uint", $list, "int", $range )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDeleteTextures( $n, ByRef $textures )
	Local $a = 0
	If IsArray( $textures ) Then
		$a = 1
		If UBound( $textures ) <> $n Or _
			UBound( $textures, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			DllCall( $dllOpenGL32, "none", "glDeleteTextures", "int", 1, "uint*", $textures[$i] )
			If @error Then Return SetError( 3, 0, 0 )
		Next
	Else
		DllCall( $dllOpenGL32, "none", "glDeleteTextures", "int", 1, "uint*", $textures )
		If @error Then Return SetError( 3, 0, 0 )
	EndIf
EndFunc

Func glDepthFunc( $func )
	DllCall( $dllOpenGL32, "none", "glDepthFunc", "uint", $func )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDepthMask( $flag )
	DllCall( $dllOpenGL32, "none", "glDepthMask", "ubyte", $flag )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDepthRange( $zNear, $zFar )
	DllCall( $dllOpenGL32, "none", "glDepthRange", "double", $zNear, "double", $zFar )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDisable( $cap )
	DllCall( $dllOpenGL32, "none", "glDisable", "uint", $cap )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDisableClientState( $cap )
	DllCall( $dllOpenGL32, "none", "glDisableClientState", "uint", $cap )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDrawArrays( $mode, $first, $count )
	DllCall( $dllOpenGL32, "none", "glDrawArrays", "uint", $mode, "int", $first, "int", $count )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDrawElements( $mode, $count, $type, $indices )
	DllCall( $dllOpenGL32, "none", "glDrawElements", "uint", $mode, "int", $count, "uint", $type, "ptr", $indices )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glEnable( $cap )
	DllCall( $dllOpenGL32, "none", "glEnable", "uint", $cap )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glEnableClientState( $cap )
	DllCall( $dllOpenGL32, "none", "glEnableClientState", "uint", $cap )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glEnd()
	DllCall( $dllOpenGL32, "none", "glEnd" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glEndList()
	DllCall( $dllOpenGL32, "none", "glEndList" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glFlush()
	DllCall( $dllOpenGL32, "none", "glFlush" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glFogf( $pname, $param )
	DllCall( $dllOpenGL32, "none", "glFogf", "uint", $pname, "float", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glFogfv( $pname, ByRef $params )
	If IsPtr( $params ) Then
		DllCall( $dllOpenGL32, "none", "glFogfv", "uint", $pname, "ptr", $params )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $params ) And UBound( $params, 0 ) = 1 Then
		Local $n = UBound( $params ), $stFloat
		If Not ( $pname = $GL_FOG_COLOR And $n = 4 ) Or _
					 ( $pname <> $GL_FOG_COLOR And $n = 1 ) Then _
						 Return SetError( 1, 0, 0 )
		$stFloat = DllStructCreate( "float[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $stFloat, 1, $params[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glFogfv", "uint", $pname, "struct*", $stFloat )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glFogi( $pname, $param )
	DllCall( $dllOpenGL32, "none", "glFogi", "uint", $pname, "int", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glFrontFace( $mode )
	DllCall( $dllOpenGL32, "none", "glFrontFace", "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glGenLists( $range )
	Local $aRet = DllCall( $dllOpenGL32, "uint", "glGenLists", "int", $range )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glGenTextures( $n, ByRef $textures )
	Local $a = 0
	If IsArray( $textures ) Then
		$a = 1
		If UBound( $textures ) <> $n Or _
			UBound( $textures, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			Local $aRet = DllCall( $dllOpenGL32, "none", "glGenTextures", "int", 1, "uint*", $textures )
			If @error Then Return SetError( 3, 0, 0 )
			$textures[$i] = $aRet[2]
		Next
	Else
		Local $aRet = DllCall( $dllOpenGL32, "none", "glGenTextures", "int", 1, "uint*", $textures )
		If @error Then Return SetError( 3, 0, 0 )
		$textures = $aRet[2]
	EndIf
EndFunc

Func glGetError()
	Local $aRet = DllCall( $dllOpenGL32, "uint", "glGetError" )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glGetFloatv( $pname, ByRef $params )
	Local $a = 0, $n = 1
	If IsArray( $params ) Then
		$a = 1
		$n = UBound( $params )
		If UBound( $params, 0 ) <> 1 Then _
			Return SetError( 1, 0, 0 )
	EndIf
	Local $stFloat = DllStructCreate( "float[" & $n & "]" )
	DllCall( $dllOpenGL32, "none", "glGetFloatv", "uint", $pname, "struct*", $stFloat )
	If @error Then Return SetError( 3, 0, 0 )
	If $a Then
		For $i = 0 To $n - 1
			$params[$i] = DllStructGetData( $stFloat, 1, $i+1 )
		Next
	Else
		$params = DllStructGetData( $stFloat, 1, 1 )
	EndIf
EndFunc

Func glGetIntegerv( $pname, ByRef $params )
	Local $a = 0, $n = 1
	If IsArray( $params ) Then
		$a = 1
		$n = UBound( $params )
		If UBound( $params, 0 ) <> 1 Then _
			Return SetError( 1, 0, 0 )
	EndIf
	Local $stInt = DllStructCreate( "int[" & $n & "]" )
	DllCall( $dllOpenGL32, "none", "glGetIntegerv", "uint", $pname, "struct*", $stInt )
	If @error Then Return SetError( 3, 0, 0 )
	If $a Then
		For $i = 0 To $n - 1
			$params[$i] = DllStructGetData( $stInt, 1, $i+1 )
		Next
	Else
		$params = DllStructGetData( $stInt, 1, 1 )
	EndIf
EndFunc

Func glGetMaterialfv( $face, $pname, ByRef $params )
	Local $size
	Switch $pname
		Case $GL_SHININESS
			$size = 1
		Case $GL_COLOR_INDEXES
			$size = 3
		Case $GL_AMBIENT, $GL_DIFFUSE, $GL_SPECULAR, $GL_EMISSION
			$size = 4
		Case Else
			Return SetError( 1, 0, 0 )
	EndSwitch
	Local $stParams = DllStructCreate( "float[" & $size & "]" )
	DllCall( $dllOpenGL32, "none", "glGetMaterialfv", "uint", $face, "uint", $pname, "struct*", $stParams )
	If @error Then Return SetError( 3, 0, 0 )
	Local $aParams[$size]
	For $i = 0 To $size - 1
		$aParams[$i] = DllStructGetData( $stParams, 1, $i+1 )
	Next
	$params = $aParams
EndFunc

Func glGetMaterialiv( $face, $pname, ByRef $params )
	Local $size
	Switch $pname
		Case $GL_SHININESS
			$size = 1
		Case $GL_COLOR_INDEXES
			$size = 3
		Case $GL_AMBIENT, $GL_DIFFUSE, $GL_SPECULAR, $GL_EMISSION
			$size = 4
		Case Else
			Return SetError( 1, 0, 0 )
	EndSwitch
	Local $stParams = DllStructCreate( "int[" & $size & "]" )
	DllCall( $dllOpenGL32, "none", "glGetMaterialiv", "uint", $face, "uint", $pname, "struct*", $stParams )
	If @error Then Return SetError( 3, 0, 0 )
	Local $aParams[$size]
	For $i = 0 To $size - 1
		$aParams[$i] = DllStructGetData( $stParams, 1, $i+1 )
	Next
	$params = $aParams
EndFunc

Func glGetString( $name )
	Local $aRet = DllCall( $dllOpenGL32, "str", "glGetString", "uint", $name )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glGetTexImage( $target, $level, $format, $type, ByRef $pixels )
	If Not IsArray( $pixels ) Or UBound( $pixels, 0 ) <> 1 Then Return SetError( 1, 0, 0 )
	Local $sType, $n = UBound( $pixels ), $struct
	Switch $type
		Case $GL_BYTE, $GL_UNSIGNED_BYTE
			$sType = "byte"
		Case $GL_SHORT
			$sType = "short"
		Case $GL_UNSIGNED_SHORT
			$sType = "ushort"
		Case $GL_INT
			$sType = "int"
		Case $GL_UNSIGNED_INT
			$sType = "uint"
		Case $GL_FLOAT
			$sType = "float"
		Case Else
			Return SetError( 1, 0, 0 )
	EndSwitch
	$struct = DllStructCreate( $sType & "[" & $n & "]" )
	DllCall( $dllOpenGL32, "none", "glGetTexImage", "uint", $target, "int", $level, "uint", $format, "uint", $type, "struct*", $struct )
	If @error Then Return SetError( 3, 0, 0 )
	For $i = 0 To $n - 1
		$pixels[$i] = DllStructGetData( $struct, 1, $i+1 )
	Next
EndFunc

Func glHint( $target, $mode )
	DllCall( $dllOpenGL32, "none", "glHint", "uint", $target, "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glIsEnabled( $cap )
	Local $aRet = DllCall( $dllOpenGL32, "ubyte", "glIsEnabled", "uint", $cap )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glLightf( $light, $pname, $param )
	DllCall( $dllOpenGL32, "none", "glLightf", "uint", $light, "uint", $pname, "float", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glLightfv( $light, $pname, ByRef $params )
	If IsPtr( $params ) Then
		DllCall( $dllOpenGL32, "none", "glLightfv", "uint", $light, "uint", $pname, "ptr", $params )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $params ) And UBound( $params, 0 ) = 1 Then
		Local $n = UBound( $params )
		Local $stFloat = DllStructCreate( "float[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $stFloat, 1, $params[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glLightfv", "uint", $light, "uint", $pname, "struct*", $stFloat )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glLineWidth( $width )
	DllCall( $dllOpenGL32, "none", "glLineWidth", "float", $width )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glListBase( $base )
	DllCall( $dllOpenGL32, "none", "glListBase", "uint", $base )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glLoadIdentity()
	DllCall( $dllOpenGL32, "none", "glLoadIdentity" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glLoadMatrixf( $m )
	If IsArray( $m ) And UBound( $m, 0 ) = 1 And UBound( $m ) = 16 Then
		Local $struct = DllStructCreate( "float[16]" )
		For $i = 0 To 15
			DllStructSetData( $struct, 1, $m[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glLoadMatrixf", "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glMaterialf( $face, $pname, $param )
	If $pname <> $GL_SHININESS Then Return SetError( 1, 0, 0 )
	DllCall( $dllOpenGL32, "none", "glMaterialf", "uint", $face, "uint", $pname, "float", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glMaterialfv( $face, $pname, ByRef $params )
	If IsPtr( $params ) Then
		DllCall( $dllOpenGL32, "none", "glMaterialfv", "uint", $face, "uint", $pname, "ptr", $params )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $params ) And UBound( $params, 0 ) = 1 Then
		Local $size = UBound( $params )
		Switch $size
			Case 1
				If $pname <> $GL_SHININESS Then _
					Return SetError( 1, 0, 0 )
			Case 3
				If $pname <> $GL_COLOR_INDEXES Then _
					Return SetError( 1, 0, 0 )
			Case 4
				If $pname = $GL_SHININESS Or _
					 $pname = $GL_COLOR_INDEXES Then _
						 Return SetError( 1, 0, 0 )
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		Local $stParams = DllStructCreate( "float[" & $size & "]" )
		For $i = 0 To $size
			DllStructSetData( $stParams, 1, $params[$i], $i+1 )
		Next
		DllCall( $dllOpenGL32, "none", "glMaterialfv", "uint", $face, "uint", $pname, "struct*", $stParams )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glMateriali( $face, $pname, $param )
	If $pname <> $GL_SHININESS Then Return SetError( 1, 0, 0 )
	DllCall( $dllOpenGL32, "none", "glMaterialf", "uint", $face, "uint", $pname, "int", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glMaterialiv( $face, $pname, ByRef $params )
	If IsPtr( $params ) Then
		DllCall( $dllOpenGL32, "none", "glMaterialiv", "uint", $face, "uint", $pname, "ptr", $params )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $params ) And UBound( $params, 0 ) = 1 Then
		Local $size = UBound( $params )
		Switch $size
			Case 1
				If $pname <> $GL_SHININESS Then _
					Return SetError( 1, 0, 0 )
			Case 3
				If $pname <> $GL_COLOR_INDEXES Then _
					Return SetError( 1, 0, 0 )
			Case 4
				If $pname = $GL_SHININESS Or _
					 $pname = $GL_COLOR_INDEXES Then _
						 Return SetError( 1, 0, 0 )
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		Local $stParams = DllStructCreate( "int[" & $size & "]" )
		For $i = 0 To $size
			DllStructSetData( $stParams, 1, $params[$i], $i+1 )
		Next
		DllCall( $dllOpenGL32, "none", "glMaterialiv", "uint", $face, "uint", $pname, "struct*", $stParams )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glMatrixMode( $mode )
	DllCall( $dllOpenGL32, "none", "glMatrixMode", "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glMultMatrixd( $m )
	Local $n = 16
	If Not IsArray( $m ) Or UBound( $m ) <> $n Or UBound( $m, 0 ) <> 1 Then Return SetError( 1, 0, 0 )
	Local $tDouble = DllStructCreate( "double[16]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tDouble, 1, $m[$i], $i+1 )
	Next
	DllCall( $dllOpenGL32, "none", "glMultMatrixd", "struct*", $tDouble )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glMultMatrixf( $m )
	Local $n = 16
	If Not IsArray( $m ) Or UBound( $m ) <> $n Or UBound( $m, 0 ) <> 1 Then Return SetError( 1, 0, 0 )
	Local $tFloat = DllStructCreate( "float[16]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tFloat, 1, $m[$i], $i+1 )
	Next
	DllCall( $dllOpenGL32, "none", "glMultMatrixf", "struct*", $tFloat )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glNewList( $list, $mode )
	DllCall( $dllOpenGL32, "none", "glNewList", "uint", $list, "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glNormal3f( $nx, $ny, $nz )
	DllCall( $dllOpenGL32, "none", "glNormal3f", "float", $nx, "float", $ny, "float", $nz )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glNormalPointer( $type, $stride, $pointer )
	If IsPtr( $pointer ) Then
		DllCall( $dllOpenGL32, "none", "glNormalPointer", "uint", $type, "int", $stride, "ptr", $pointer )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $pointer ) And UBound( $pointer, 0 ) = 1 Then
		Local $sType, $n = UBound( $pointer )
		Local Static $struct
		Switch $type
			Case $GL_BYTE
				$sType = "byte"
			Case $GL_SHORT
				$sType = "short"
			Case $GL_INT
				$sType = "int"
			Case $GL_FLOAT
				$sType = "float"
			Case $GL_DOUBLE
				$sType = "double"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		$struct = DllStructCreate( $sType & "[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $struct, 1, $pointer[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glNormalPointer", "uint", $type, "int", $stride, "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glOrtho( $left, $right, $bottom, $top, $zNear, $zFar )
	DllCall( $dllOpenGL32, "none", "glOrtho", "double", $left, "double", $right, "double", $bottom, "double", $top, "double", $zNear, "double", $zFar )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPointSize( $size )
	DllCall( $dllOpenGL32, "none", "glPointSize", "float", $size )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPolygonMode( $face, $mode )
	DllCall( $dllOpenGL32, "none", "glPolygonMode", "uint", $face, "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPolygonOffset( $factor, $units )
	DllCall( $dllOpenGL32, "none", "glPolygonOffset", "float", $factor, "float", $units )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPopAttrib()
	DllCall( $dllOpenGL32, "none", "glPopAttrib" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPopMatrix()
	DllCall( $dllOpenGL32, "none", "glPopMatrix" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPushAttrib( $mask )
	DllCall( $dllOpenGL32, "none", "glPushAttrib", "uint", $mask )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glPushMatrix()
	DllCall( $dllOpenGL32, "none", "glPushMatrix" )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glRasterPos2f( $x, $y )
	DllCall( $dllOpenGL32, "none", "glRasterPos2f", "float", $x, "float", $y )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glReadPixels( $x, $y, $width, $height, $format, $type, $pixels )
	DllCall( $dllOpenGL32, "none", "glReadPixels", "int", $x, "int", $y, "int", $width, "int", $height, "uint", $format, "uint", $type, "struct*", $pixels )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glRotatef( $angle, $x, $y, $z )
	DllCall( $dllOpenGL32, "none", "glRotatef", "float", $angle, "float", $x, "float", $y, "float", $z )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glScalef( $x, $y, $z )
	DllCall( $dllOpenGL32, "none", "glScalef", "float", $x, "float", $y, "float", $z )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glScissor( $x, $y, $width, $height )
	DllCall( $dllOpenGL32, "none", "glScissor", "int", $x, "int", $y, "int", $width, "int", $height )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glShadeModel( $mode )
	DllCall( $dllOpenGL32, "none", "glShadeModel", "uint", $mode )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexCoord2d( $s, $t )
	DllCall( $dllOpenGL32, "none", "glTexCoord2d", "double", $s, "double", $t )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexCoord2f( $s, $t )
	DllCall( $dllOpenGL32, "none", "glTexCoord2f", "float", $s, "float", $t )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexCoord3f( $s, $t, $r )
	DllCall( $dllOpenGL32, "none", "glTexCoord3f", "float", $s, "float", $t, "float", $r )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexCoordPointer( $size, $type, $stride, $pointer )
	If IsPtr( $pointer ) Then
		DllCall( $dllOpenGL32, "none", "glTexCoordPointer", "int", $size, "uint", $type, "int", $stride, "ptr", $pointer )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $pointer ) And UBound( $pointer, 0 ) = 1 Then
		Local $sType, $n = UBound( $pointer )
		Local Static $struct
		Switch $type
			Case $GL_SHORT
				$sType = "short"
			Case $GL_INT
				$sType = "int"
			Case $GL_FLOAT
				$sType = "float"
			Case $GL_DOUBLE
				$sType = "double"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		$struct = DllStructCreate( $sType & "[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $struct, 1, $pointer[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glTexCoordPointer", "int", $size, "uint", $type, "int", $stride, "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glTexEnvf( $target, $pname, $param )
	DllCall( $dllOpenGL32, "none", "glTexEnvf", "uint", $target, "uint", $pname, "float", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexImage1D( $target, $level, $internalformat, $width, $border, $format, $type, $pixels )
	DllCall( $dllOpenGL32, "none", "glTexImage1D", "uint", $target, "int", $level, "int", $internalformat, "int", $width, "int", $border, "uint", $format, "uint", $type, "ptr", $pixels )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexImage2D( $target, $level, $internalFormat, $width, $height, $border, $format, $type, $pixels )
	DllCall( $dllOpenGL32, "none", "glTexImage2D", "uint", $target, "int", $level, "int", $internalFormat, "int", $width, "int", $height, "int", $border, "uint", $format, "uint", $type, "ptr", $pixels )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexparameterf( $target, $pname, $param )
	DllCall( $dllOpenGL32, "none", "glTexparameterf", "uint", $target, "uint", $pname, "float", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTexparameteri( $target, $pname, $param )
	DllCall( $dllOpenGL32, "none", "glTexparameteri", "uint", $target, "uint", $pname, "int", $param )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glTranslatef( $x, $y, $z )
	DllCall( $dllOpenGL32, "none", "glTranslatef", "float", $x, "float", $y, "float", $z )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertex2d( $x, $y )
	DllCall( $dllOpenGL32, "none", "glVertex2d", "double", $x, "double", $y )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertex2f( $x, $y )
	DllCall( $dllOpenGL32, "none", "glVertex2f", "float", $x, "float", $y )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertex2i( $x, $y )
	DllCall( $dllOpenGL32, "none", "glVertex2i", "int", $x, "int", $y )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertex3f( $x, $y, $z )
	DllCall( $dllOpenGL32, "none", "glVertex3f", "float", $x, "float", $y, "float", $z )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertex3fv( ByRef $v )
	If IsPtr( $v ) Then
		DllCall( $dllOpenGL32, "none", "glVertex3fv", "ptr", $v )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $v ) And UBound( $v, 0 ) = 1 And UBound( $v ) = 3 Then
		Local $stVertex = DllStructCreate( "float[3]" )
		For $i = 0 To 2
			DllStructSetData( $stVertex, 1, $v[$i], $i+1 )
		Next
		DllCall( $dllOpenGL32, "none", "glVertex3fv", "struct*", $stVertex )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glVertex3i( $x, $y, $z )
	DllCall( $dllOpenGL32, "none", "glVertex3i", "int", $x, "int", $y, "int", $z )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertexPointer( $size, $type, $stride, $pointer )
	If IsPtr( $pointer ) Then
		DllCall( $dllOpenGL32, "none", "glVertexPointer", "int", $size, "uint", $type, "int", $stride, "ptr", $pointer )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $pointer ) And UBound( $pointer, 0 ) = 1 Then
		Local $sType, $n = UBound( $pointer )
		Local Static $struct
		Switch $type
			Case $GL_SHORT
				$sType = "short"
			Case $GL_INT
				$sType = "int"
			Case $GL_FLOAT
				$sType = "float"
			Case $GL_DOUBLE
				$sType = "double"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		$struct = DllStructCreate( $sType & "[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $struct, 1, $pointer[$i], $i + 1 )
		Next
		DllCall( $dllOpenGL32, "none", "glVertexPointer", "int", $size, "uint", $type, "int", $stride, "struct*", $struct )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glViewport( $x, $y, $width, $height )
	DllCall( $dllOpenGL32, "none", "glViewport", "int", $x, "int", $y, "int", $width, "int", $height )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ---------------- OpenGL Utility Library functions (GLU) -----------------

Func gluBuild1DMipmaps( $target, $components, $width, $format, $type, $data )
	Local $aRet = DllCall( $dllGLU32, "int", "gluBuild1DMipmaps", "uint", $target, "int", $components, "int", $width, "uint", $format, "uint", $type, "ptr", $data )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func gluBuild2DMipmaps( $target, $components, $width, $height, $format, $type, $data )
	Local $aRet = DllCall( $dllGLU32, "int", "gluBuild2DMipmaps", "uint", $target, "int", $components, "int", $width, "int", $height, "uint", $format, "uint", $type, "ptr", $data )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func gluDeleteQuadric( $quad )
	DllCall( $dllGLU32, "none", "gluDeleteQuadric", "ptr", $quad )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func gluErrorString( $errCode )
	Local $aRet = DllCall( $dllGLU32, "ptr", "gluErrorString", "uint", $errCode )
	If @error Then Return SetError( 3, 0, 0 )
	Local $stString = DllStructCreate( "char[512]", $aRet[0] )
	Return DllStructGetData( $stString, 1 )
EndFunc

Func gluLookAt( $eyex, $eyey, $eyez, $centerx, $centery, $centerz, $upx, $upy, $upz )
	DllCall( $dllGLU32, "none", "gluLookAt", _
											"double", $eyex,    "double", $eyey,    "double", $eyez, _
											"double", $centerx, "double", $centery, "double", $centerz, _
											"double", $upx,     "double", $upy,     "double", $upz )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func gluNewQuadric()
	Local $aRet = DllCall( $dllGLU32, "ptr", "gluNewQuadric" )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func gluPerspective( $fovy, $aspect, $zNear, $zFar )
	DllCall( $dllGLU32, "none", "gluPerspective", "double", $fovy, "double", $aspect, "double", $zNear, "double", $zFar )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func gluProject( $objx, $objy, $objz, ByRef $modelMatrix, ByRef $projMatrix, ByRef $viewport, ByRef $winx, ByRef $winy, ByRef $winz )
	Local $n
	$n = UBound( $modelMatrix )
	Local $tModelMatrix = DllStructCreate( "double[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tModelMatrix, 1, $modelMatrix[$i], $i+1 )
	Next
	$n = UBound( $projMatrix )
	Local $tProjMatrix = DllStructCreate( "double[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tProjMatrix, 1, $projMatrix[$i], $i+1 )
	Next
	$n = UBound( $viewport )
	Local $tViewport = DllStructCreate( "int[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $tViewport, 1, $viewport[$i], $i+1 )
	Next
	Local $aRet = DllCall( $dllGLU32, "int", "gluProject", _
																		"double", $objx, "double", $objy, "double", $objz, _
																		"struct*", $tModelMatrix, "struct*", $tProjMatrix, "struct*", $tViewport, _
																		"double*", 0, "double*", 0, "double*", 0 ) 
	If @error Then Return SetError( 3, 0, 0 )
	$winx = $aRet[7]
	$winy = $aRet[8]
	$winz = $aRet[9]
EndFunc

Func gluQuadricDrawStyle( $qobj, $drawStyle )
	DllCall( $dllGLU32, "none", "gluQuadricDrawStyle", "ptr", $qobj, "uint", $drawStyle )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func gluQuadricTexture( $qobj, $textureCoords )
	DllCall( $dllGLU32, "none", "gluQuadricTexture", "ptr", $qobj, "byte", $textureCoords )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func gluSphere( $qobj, $radius, $slices, $stacks )
	DllCall( $dllGLU32, "none", "gluSphere", "ptr", $qobj, "double", $radius, "int", $slices, "int", $stacks )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ------------------------------ OpenGL 1.2 ------------------------------


; ----------------------------- OpenGL 1.2.1 -----------------------------


; ------------------------------ OpenGL 1.3 ------------------------------

Func glActiveTexture( $texture )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glActiveTexture" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $texture )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glClientActiveTexture( $texture )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glClientActiveTexture" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $texture )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ------------------------------ OpenGL 1.4 ------------------------------


; ------------------------------ OpenGL 1.5 ------------------------------

Func glBindBuffer( $target, $buffer )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glBindBuffer" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $target , "uint", $buffer )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glBufferData( $target, $type, $size, $data, $usage )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glBufferData" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If IsPtr( $data ) Then
		DllCallAddress( "none", $ptr, "uint", $target, "int", $size, "ptr", $data, "uint", $usage )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $data ) And UBound( $data, 0 ) = 1 Then
		Switch $type
			Case $GL_BYTE, $GL_UNSIGNED_BYTE
				$type = "byte"
			Case $GL_SHORT
				$type = "short"
			Case $GL_UNSIGNED_SHORT
				$type = "ushort"
			Case $GL_INT
				$type = "int"
			Case $GL_UNSIGNED_INT
				$type = "uint"
			Case $GL_FLOAT
				$type = "float"
			Case Else
				Return SetError( 1, 0, 0 )
		EndSwitch
		Local $n = UBound( $data ), $struct
		$struct = DllStructCreate( $type & "[" & $n & "]" )
		For $i = 0 To $n - 1
			DllStructSetData( $struct, 1, $data[$i], $i + 1 )
		Next
		DllCallAddress( "none", $ptr, "uint", $target, "int", $size, "struct*", $struct, "uint", $usage )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf Not IsArray( $data ) And $data = 0 Then
		DllCallAddress( "none", $ptr, "uint", $target, "int", $size, "ptr", 0, "uint", $usage )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glBufferSubData( $target, $type, $offset, $size, $data )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glBufferSubData" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If Not IsArray( $data ) Or _
		UBound( $data, 0 ) <> 1 Then _
			Return SetError( 1, 0, 0 )
	Switch $type
		Case $GL_BYTE, $GL_UNSIGNED_BYTE
			$type = "byte"
		Case $GL_SHORT
			$type = "short"
		Case $GL_UNSIGNED_SHORT
			$type = "ushort"
		Case $GL_INT
			$type = "int"
		Case $GL_UNSIGNED_INT
			$type = "uint"
		Case $GL_FLOAT
			$type = "float"
		Case Else
			Return SetError( 1, 0, 0 )
	EndSwitch
	Local $n = UBound( $data ), $struct
	$struct = DllStructCreate( $type & "[" & $n & "]" )
	For $i = 0 To $n - 1
		DllStructSetData( $struct, 1, $data[$i], $i + 1 )
	Next
	DllCallAddress( "none", $ptr, "uint", $target, "int", $offset, "int", $size, "struct*", $struct )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDeleteBuffers( $n, ByRef $buffers )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDeleteBuffers" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $a = 0
	If IsArray( $buffers ) Then
		$a = 1
		If UBound( $buffers ) <> $n Or _
			UBound( $buffers, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			DllCallAddress( "none", $ptr, "int", 1, "uint*", $buffers[$i] )
			If @error Then Return SetError( 3, 0, 0 )
		Next
	Else
		DllCallAddress( "none", $ptr, "int", 1, "uint*", $buffers )
		If @error Then Return SetError( 3, 0, 0 )
	EndIf
EndFunc

Func glGenBuffers( $n, ByRef $buffers )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGenBuffers" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $a = 0, $buffer
	If IsArray( $buffers ) Then
		$a = 1
		If UBound( $buffers ) <> $n Or _
			UBound( $buffers, 0 ) <> 1 Then _
				Return SetError( 1, 0, 0 )
	EndIf
	If $a Then
		For $i = 0 To $n - 1
			Local $aRet = DllCallAddress( "none", $ptr, "int", 1, "uint*", $buffer )
			If @error Then Return SetError( 3, 0, 0 )
			$buffers[$i] = $aRet[2]
		Next
	Else
		Local $aRet = DllCallAddress( "none", $ptr, "int", 1, "uint*", $buffer )
		If @error Then Return SetError( 3, 0, 0 )
		$buffers = $aRet[2]
	EndIf
EndFunc

Func glGetBufferParameteriv( $target, $value, ByRef $data )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetBufferParameteriv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $target, "uint", $value, "int*", $data )
	If @error Then Return SetError( 3, 0, 0 )
	$data = $aRet[3]
EndFunc

Func glGetBufferSubData( $target, $type, $offset, $size, ByRef $data )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetBufferSubData" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $iBytes
	Switch $type
		Case $GL_BYTE, $GL_UNSIGNED_BYTE
			$type = "byte"
			$iBytes = 1
		Case $GL_SHORT
			$type = "short"
			$iBytes = 2
		Case $GL_UNSIGNED_SHORT
			$type = "ushort"
			$iBytes = 2
		Case $GL_INT
			$type = "int"
			$iBytes = 4
		Case $GL_UNSIGNED_INT
			$type = "uint"
			$iBytes = 4
		Case $GL_FLOAT
			$type = "float"
			$iBytes = 4
		Case Else
			Return SetError( 1, 0, 0 )
	EndSwitch
	If Mod( $size, $iBytes ) Then _
		Return SetError( 1, 0, 0 )
	Local $struct = DllStructCreate( $type & "[" & $size / $iBytes & "]" )
	DllCallAddress( "none", $ptr, "uint", $target, "int", $offset, "int", $size, "struct*", $struct )
	If @error Then Return SetError( 3, 0, 0 )
	$data = $struct
EndFunc

; ------------------------------ OpenGL 2.0 ------------------------------

Func glAttachShader( $program, $shader )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glAttachShader" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program, "uint", $shader )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glCompileShader( $shader )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glCompileShader" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $shader )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glCreateProgram()
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glCreateProgram" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "uint", $ptr )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glCreateShader( $type )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glCreateShader" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "uint", $ptr, "uint", $type )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glDeleteProgram( $program )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDeleteProgram" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDeleteShader( $shader )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDeleteShader" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $shader )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDetachShader( $program, $shader )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDetachShader" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program, "uint", $shader )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glDisableVertexAttribArray( $uint )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glDisableVertexAttribArray" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $uint )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glEnableVertexAttribArray( $index )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glEnableVertexAttribArray" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $index )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glGetAttribLocation( $program, $name )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetAttribLocation" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "int", $ptr, "uint", $program, "str", $name )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glGetProgramiv( $program, $pname, ByRef $param )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetProgramiv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $program, "uint", $pname, "int*", $param )
	If @error Then Return SetError( 3, 0, 0 )
	$param = $aRet[3]
EndFunc

Func glGetShaderiv( $shader, $pname, ByRef $param )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetShaderiv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $shader, "uint", $pname, "int*", $param )
	If @error Then Return SetError( 3, 0, 0 )
	$param = $aRet[3]
EndFunc

Func glGetShaderSource( $obj, $maxLength, ByRef $length, ByRef $source )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetShaderSource" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "none", $ptr, "uint", $obj, "int", $maxLength, "int*", $length, "str", $source )
	If @error Then Return SetError( 3, 0, 0 )
	$length = $aRet[3]
	$source = $aRet[4]
EndFunc

Func glGetUniformLocation( $program, $name )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glGetUniformLocation" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	Local $aRet = DllCallAddress( "int", $ptr, "uint", $program, "str", $name )
	If @error Then Return SetError( 3, 0, 0 )
	Return $aRet[0]
EndFunc

Func glLinkProgram( $program )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glLinkProgram" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glShaderSource( $shader, $count, $strings, $lengths )
	Local $stShader = DllStructCreate( "char[" & $lengths & "]" )
	DllStructSetData( $stShader, 1, $strings )
	Local $pStrings = DllStructGetPtr( $stShader )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glShaderSource" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $shader, "int", $count, "ptr*", $pStrings, "int*", $lengths )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform1f( $location, $v0 )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform1f" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "int", $location, "float", $v0 )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform1i( $location, $v0 )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform1i" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "int", $location, "int", $v0 )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform2f( $location, $v0, $v1 )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform2f" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "int", $location, "float", $v0, "float", $v1 )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform3f( $location, $v0, $v1, $v2 )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform3f" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "int", $location, "float", $v0, "float", $v1, "float", $v2 )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform4f( $location, $v0, $v1, $v2, $v3 )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform4f" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "int", $location, "float", $v0, "float", $v1, "float", $v2, "float", $v3 )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glUniform4fv( $location, $count, ByRef $value )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniform4fv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If IsPtr( $value ) Then
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "ptr", $value )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $value ) And UBound( $value ) = 4 And UBound( $value, 0 ) = 1 Then
		Local $stFloat = DllStructCreate( "float[4]" )
		For $i = 0 To 3
			DllStructSetData( $stFloat, 1, $value[$i], $i + 1 )
		Next
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "struct*", $stFloat )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glUniformMatrix3fv( $location, $count, $transpose, ByRef $value )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniformMatrix3fv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If IsPtr( $value ) Then
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "byte", $transpose, "ptr", $value )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $value ) And UBound( $value ) = 9 And UBound( $value, 0 ) = 1 Then
		Local $stFloat = DllStructCreate( "float[9]" )
		For $i = 0 To 8
			DllStructSetData( $stFloat, 1, $value[$i], $i + 1 )
		Next
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "byte", $transpose, "struct*", $stFloat )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glUniformMatrix4fv( $location, $count, $transpose, ByRef $value )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUniformMatrix4fv" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	If IsPtr( $value ) Then
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "byte", $transpose, "ptr", $value )
		If @error Then Return SetError( 3, 0, 0 )
	ElseIf IsArray( $value ) And UBound( $value ) = 16 And UBound( $value, 0 ) = 1 Then
		Local $stFloat = DllStructCreate( "float[16]" )
		For $i = 0 To 15
			DllStructSetData( $stFloat, 1, $value[$i], $i + 1 )
		Next
		DllCallAddress( "none", $ptr, "int", $location, "int", $count, "byte", $transpose, "struct*", $stFloat )
		If @error Then Return SetError( 3, 0, 0 )
	Else
		Return SetError( 1, 0, 0 )
	EndIf
EndFunc

Func glUseProgram( $program )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glUseProgram" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glValidateProgram( $program )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glValidateProgram" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $program )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

Func glVertexAttribPointer( $index, $size, $type, $normalized, $stride, $pointer )
	Local Static $ptr = 0
	If $ptr = 0 Then
		$ptr = wglGetProcAddress( "glVertexAttribPointer" )
		If @error Then Return SetError( 2, @error, 0 )
	EndIf
	DllCallAddress( "none", $ptr, "uint", $index, "int", $size, "uint", $type, "byte", $normalized, "int", $stride, "ptr", $pointer )
	If @error Then Return SetError( 3, 0, 0 )
EndFunc

; ------------------------------ OpenGL 2.1 ------------------------------


; ------------------------------ OpenGL 3.0 ------------------------------


; ------------------------------ OpenGL 3.1 ------------------------------


; ------------------------------ OpenGL 3.2 ------------------------------


; ------------------------------ OpenGL 3.3 ------------------------------


; ------------------------------ OpenGL 4.0 ------------------------------


; ------------------------------ OpenGL 4.1 ------------------------------


; ------------------------------ OpenGL 4.2 ------------------------------
