; OpenGL image loading functions
;
; LoadTextureImage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
;		Loads an image into a texture and creates mipmaps
;		Wrapper around the OpenGL image loading functions
;		Calls the proper function depending on image type (suffix)
;		Image types: BMP, GIF, ICO, JPG, PNG, RGB, RGBA, TGA, TIF and WMF
;		The image can be specified as a file name or as an internet URL ( "http://..." )
;		( Internet URL is available for BMP, GIF, ICO, JPG and WMF images )
;
; LoadImage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
;		Reads a BMP, GIF, ICO, JPG and WMF image into a byte structure
;		The image can be specified as a file name or as an internet URL ( "http://..." )
;		Calls CreateTexture()
;
; LoadGDIimage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
;		Reads a BMP, JPG, PNG and TIF image into a byte structure
;		PNG and TIF images when called through LoadTextureImage()
;		BMP, JPG, PNG and TIF images when called directly
;		Calls CreateTexture()
;
; LoadTGAimage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
;		Reads a TGA ( uncompressed or RLE-encoded ) image into a byte structure
;		Calls CreateTexture()
;
; LoadRGBimage( ByRef $texture, $imgFile, $width, $height, $fTexGlu = 0 )
;		Reads a RGB and RGBA image into a byte structure
;		Calls CreateTexture()
;
; CreateTexture( ByRef $texture, $width, $height, $format, $pImage, $fTexGlu = 0 )
;		Loads a byte structure into a texture and creates mipmaps


#include-once

#include <GDIPlus.au3>
#include "oglUtilities.au3"


; -------------------- LoadTextureImage() --------------------

; Loads an image into a texture and creates mipmaps
; Wrapper around the OpenGL image loading functions
; Calls the proper function depending on image type (suffix)
; Image types: BMP, GIF, ICO, JPG, PNG, RGB, RGBA, TGA, TIF and WMF
; The image can be specified as a file name or as an internet URL ( "http://..." )
; ( Internet URL is available for BMP, GIF, ICO, JPG and WMF images )
Func LoadTextureImage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )

	If StringLeft( $imgFile, 7 ) = "http://" Then _
		Return LoadImage( $texture, $imgFile, $width, $height, $fTexGlu )

	Local $sFormat = StringRight( $imgFile, StringLen( $imgFile ) - StringInStr( $imgFile, ".", 0, -1 ) )

	Switch $sFormat

		Case "bmp", "gif", "ico", "jpg", "wmf"
			Return LoadImage( $texture, $imgFile, $width, $height, $fTexGlu )

		Case "png", "tif"
			Return LoadGDIimage( $texture, $imgFile, $width, $height, $fTexGlu )

		Case "rgb", "rgba"
			Return LoadRGBimage( $texture, $imgFile, $width, $height, $fTexGlu )

		Case "tga"
			Return LoadTGAimage( $texture, $imgFile, $width, $height, $fTexGlu )

		Case Else
			PrintError( "Function LoadTextureImage()" & @CRLF & _
				"ERROR: """ & $sFormat & """ is not a legal image format." & @CRLF & _
				"Legal formats: BMP, GIF, ICO, JPG, PNG, RGB, RGBA, TGA, TIF, WMF" )
			Return 0

	EndSwitch

EndFunc


; -------------------- LoadImage() --------------------

; IPicture interface
; http://msdn.microsoft.com/en-us/library/windows/desktop/ms680761(v=vs.85).aspx
Global Const $sIID_IPicture = "{7BF80980-BF32-101A-8BBB-00AA00300CAB}"
Global Const $tRIID_IPicture = CLSIDFromString( $sIID_IPicture )
Global Const $dtag_IPicture = _
  "get_Handle hresult(handle*);" & _                                         ; Returns the Windows GDI handle of the picture managed within this picture object.
  "get_hPal hresult(handle*);" & _                                           ; Returns a copy of the palette currently used by the picture object.
  "get_Type hresult(short*);" & _                                            ; Returns the current type of the picture.
  "get_Width hresult(long*);" & _                                            ; Returns the current width of the picture in the picture object.
  "get_Height hresult(long*);" & _                                           ; Returns the current height of the picture in the picture object.
  "Render hresult(handle;long;long;long;long;long;long;long;long;ptr);" & _  ; Draws the specified portion of the picture onto the specified device context, positioned at the specified location.
  "set_hPal hresult(handle);" & _                                            ; Sets the current palette of the picture.
  "get_CurDC hresult(handle*);" & _                                          ; Returns the current device context into which this picture is selected.
  "SelectPicture hresult(handle;handle*;handle*);" & _                       ; Selects a bitmap picture into a given device context, returning the device context in which the picture was previously selected as well as the picture's GDI handle.
  "get_KeepOriginalFormat hresult(int*);" & _                                ; Returns the current value of the picture object's KeepOriginalFormat property.
  "put_KeepOriginalFormat hresult(int);" & _                                 ; Sets the picture object's KeepOriginalFormat property.
  "PictureChanged hresult();" & _                                            ; Notifies the picture object that its picture resource changed.
  "SaveAsFile hresult(ptr*;int;long*);" & _                                  ; Saves the picture's data into a stream in the same format that it would save itself into a file.
  "get_Attributes hresult(dword*);"                                          ; Returns the current set of the picture's bit attributes.

; Reads a BMP, GIF, ICO, JPG and WMF image into a byte structure
; The image can be specified as a file name or as an internet URL ( "http://..." )
; Calls CreateTexture() to load the byte structure into a texture
Func LoadImage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
	; http://nehe.gamedev.net/tutorial/volumetric_fog__ipicture_image_loading/18007/

	If StringLeft( $imgFile, 7 ) <> "http://" Then

		; File format
		Local $sFormat = StringRight( $imgFile, StringLen( $imgFile ) - StringInStr( $imgFile, ".", 0, -1 ) )
		Switch $sFormat
			Case "bmp", "gif", "ico", "jpg", "wmf"
			Case Else
				PrintError( "Function LoadImage()" & @CRLF & _
					"ERROR: """ & $sFormat & """ is not a legal image format." & @CRLF & _
					"Legal formats: BMP, GIF, ICO, JPG, WMF" )
				Return SetError(1,0,0)
		EndSwitch

		; File exists?
		If Not FileExists( $imgFile ) Then
			PrintError( "Function LoadImage()" & @CRLF & _
				"ERROR: File """ & $imgFile & """ does not exists." )
			Return SetError(2,0,0)
		EndIf

		$imgFile = GetFullPathName( $imgFile )

		; File exists?
		If Not FileExists( $imgFile ) Then
			PrintError( "Function LoadImage()" & @CRLF & _
				"ERROR: File """ & $imgFile & """ does not exists." )
			Return SetError(3,0,0)
		EndIf

	EndIf

	; Wide char name
	Local $wcImgFile = _WinAPI_MultiByteToWideChar( $imgFile, 0, 0x01, True ) ; 0x01 => $MB_PRECOMPOSED

	; Load image and get pointer to IPicture interface
	Local $aRet = DllCall( "OleAut32.dll", "long", "OleLoadPicturePath", "wstr", $wcImgFile, "long_ptr", 0, _
												 "dword", 0, "long", 0, "ptr", DllStructGetPtr( $tRIID_IPicture ), "ptr*", 0 )

	; Create IPicture interface object
	Local $oIPicture = ObjCreateInterface( $aRet[6], $sIID_IPicture, $dtag_IPicture )
	If @error Or Not IsObj( $oIPicture ) Then
		PrintError( "Function LoadImage()" & @CRLF & _
			"ERROR: Could not create IPicture interface." )
		Return SetError(4,0,0)
	EndIf

	; Create compatible device context
	Local $hDC = _WinAPI_CreateCompatibleDC( GetDC(0) )
	If @error Or Not $hDC Then
		PrintError( "Function LoadImage()" & @CRLF & _
			"ERROR: Could not create compatible device context." )
		Return SetError(5,0,0)
	EndIf

	; Width and height
	Local $iW, $iH
	$oIPicture.get_Width( $iW )
	$oIPicture.get_Height( $iH )
	; Width and height in pixels
	$width = MulDiv( $iW, GetDeviceCaps( GetDC(0), 88 ), 2540 )		; 88 => LOGPIXELSX
	$height = MulDiv( $iH, GetDeviceCaps( GetDC(0), 90 ), 2540 )	; 90 => LOGPIXELSY

	; Create and fill BITMAPINFO structure
	Local $tBITMAPINFO = DllStructCreate( $tagBITMAPINFO )
	DllStructSetData( $tBITMAPINFO, "biSize", DllStructGetSize( $tBITMAPINFO ) - 4 )
	DllStructSetData( $tBITMAPINFO, "biWidth", $width )
	DllStructSetData( $tBITMAPINFO, "biHeight", $height )
	DllStructSetData( $tBITMAPINFO, "biPlanes", 1 )
	DllStructSetData( $tBITMAPINFO, "biBitCount", 24 )
	DllStructSetData( $tBITMAPINFO, "biCompression", 0 ) ; 0 => $BI_RGB

	; Create bitmap object
	Local $hBitmapObj, $pBits
	$hBitmapObj = CreateDIBSection( $hDC, $tBITMAPINFO, 0, $pBits ) ; 0 => $DIB_RGB_COLORS
	If @error Or Not $hBitmapObj Or Not IsPtr( $pBits ) Then
		PrintError( "Function LoadImage()" & @CRLF & _
			"ERROR: Could not create bitmap object for """ & $imgFile & """." )
		Return SetError(6,0,0)
	EndIf

	; Select handle to device context and bitmap object
	_WinAPI_SelectObject( $hDC, $hBitmapObj )

	; Render image into bitmap object
	If $oIPicture.Render( $hDC, 0, 0, $width, $height, 0, $iH, $iW, -$iH, 0 ) <> 0 Then
		PrintError( "Function LoadImage()" & @CRLF & _
			"ERROR: Could not render image """ & $imgFile & """ into bitmap object." )
		Return SetError(7,0,0)
	EndIf

	; Create byte structure
	Local $tPixels = DllStructCreate( "byte[" & $width * $height * 3 & "]", $pBits )
	Local $pImage = DllStructGetPtr( $tPixels )

	Local $iRet = 1

	; Create texture
	If Not CreateTexture( $texture, $width, $height, $GL_BGR, $pImage, $fTexGlu ) Then
		PrintTextureError( "LoadImage()", @error, @extended )
		$iRet = 0
	EndIf

	; Delete resources
	_WinAPI_DeleteObject( $hBitmapObj )
	_WinAPI_DeleteDC( $hDC )

	Return $iRet
EndFunc

; Copied from AutoItObject.au3 by
; the AutoItObject team / Prog@ndy.
Func CLSIDFromString( $sString )
	Local $tCLSID = DllStructCreate( "dword;word;word;byte[8]" )
	Local $aRet = DllCall( "Ole32.dll", "long", "CLSIDFromString", "wstr", $sString, "ptr", DllStructGetPtr( $tCLSID ) )
	If @error Then Return SetError( 1, @error, 0 )
	If $aRet[0] <> 0 Then Return SetError( 2, $aRet[0], 0 )
	Return $tCLSID
EndFunc

; Copied from WinAPIEx.au3/WinAPIGdi.au3
Func CreateDIBSection($hDC, $tBITMAPINFO, $iUsage, ByRef $pBits, $hSection = 0, $iOffset = 0)

	$pBits = 0

	Local $Ret = DllCall($dllGDI32, 'ptr', 'CreateDIBSection', 'hwnd', $hDC, 'struct*', $tBITMAPINFO, 'uint', $iUsage, 'ptr*', 0, 'ptr', $hSection, 'dword', $iOffset)

	If (@error) Or (Not $Ret[0]) Then
		Return SetError(1, 0, 0)
	EndIf
	$pBits = $Ret[4]
	Return $Ret[0]
EndFunc

; Copied from WinAPIEx.au3
Func GetFullPathName($sFile)

	Local $Ret = DllCall($dllKernel32, 'dword', 'GetFullPathNameW', 'wstr', $sFile, 'dword', 4096, 'wstr', '', 'ptr', 0)

	If (@error) Or (Not $Ret[0]) Then
		Return SetError(1, 0, '')
	EndIf
	Return $Ret[3]
EndFunc

Func MulDiv( $nNumber, $nNumerator, $nDenominator )
	Local $aRet = DllCall( $dllKernel32, "int", "MulDiv", "int", $nNumber, "int", $nNumerator, "int", $nDenominator )
	Return $aRet[0]
EndFunc

Func GetDeviceCaps( $hdc, $nIndex )
	Local $aRet = DllCall( $dllGDI32, "int", "GetDeviceCaps", "ptr", $hdc, "int", $nIndex )
	Return $aRet[0]
EndFunc


; -------------------- LoadGDIimage() --------------------

; Reads a BMP, JPG, PNG and TIF image into a byte structure
; Calls CreateTexture() to load the byte structure into a texture
Func LoadGDIimage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )
	Local $bytes, $format, $size

	; File format
	Local $sFormat = StringRight( $imgFile, StringLen( $imgFile ) - StringInStr( $imgFile, ".", 0, -1 ) )
	Switch $sFormat
		Case "bmp", "jpg", "png", "tif"
		Case Else
			PrintError( "Function LoadGDIimage()" & @CRLF & _
				"ERROR: """ & $sFormat & """ is not a legal image format." & @CRLF & _
				"Legal formats: BMP, JPG, PNG, TIF" )
			Return SetError(1,0,0)
	EndSwitch

	; File exists?
	If Not FileExists( $imgFile ) Then
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: File """ & $imgFile & """ does not exists." )
		Return SetError(2,0,0)
	EndIf

	; Start GDIPlus
	If Not _GDIPlus_Startup() Then
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: Could not start GDIPlus." )
		Return SetError(3,0,0)
	EndIf

	Local $hBitmapObj, $iPixelFormat

	; Read image into bitmap
	$hBitmapObj = _GDIPlus_ImageLoadFromFile( $imgFile )
	If Not $hBitmapObj Then
		_GDIPlus_Shutdown()
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: Could not create bitmap for """ & $imgFile & """." )
		Return SetError(4,0,0)
	EndIf

	; Check pixel format
	Local $aPixelFormat = _GDIPlus_ImageGetPixelFormat( $hBitmapObj ), $iError = @error
	If Not $iError And ( $aPixelFormat[0] = $GDIP_PXF24RGB Or _
											 $aPixelFormat[0] = $GDIP_PXF32RGB Or _
											 $aPixelFormat[0] = $GDIP_PXF32ARGB ) Then
		$iPixelFormat = $aPixelFormat[0]
	ElseIf $iError Then
		_GDIPlus_Shutdown()
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: Could not get pixel format for """ & $imgFile & """." )
		Return SetError(5,0,0)
	ElseIf $aPixelFormat[0] Then
		_GDIPlus_Shutdown()
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: """ & $aPixelFormat[1] & """ is not a legal pixel format for """ & $imgFile & """." & @CRLF & _
			"Legal formats: 24 Bpp RGB, 32 Bpp RGB, 32 Bpp ARGB" )
		Return SetError(6,0,0)
	EndIf

	; Width, height
  $width = _GDIPlus_ImageGetWidth( $hBitmapObj )
	$height = _GDIPlus_ImageGetHeight( $hBitmapObj )

	; Bytes per pixel, type
	Switch $iPixelFormat
		Case $GDIP_PXF24RGB
			$bytes = 3
			$format = $GL_BGR
		Case $GDIP_PXF32RGB
			$bytes = 4
			$format = $GL_BGRA
		Case $GDIP_PXF32ARGB
			$bytes = 4
			$format = $GL_BGRA
	EndSwitch

	; Size
	$size = $width * $height * $bytes

	; Lock bitmap for reading
	Local $tBitmapData = _GDIPlus_BitmapLockBits( $hBitmapObj, 0, 0, $width, $height, $GDIP_ILMREAD, $iPixelFormat )
	If @error Then
		_GDIPlus_Shutdown()
		PrintError( "Function LoadGDIimage()" & @CRLF & _
			"ERROR: Could not lock bitmap (""" & $imgFile & """) for reading." )
		Return SetError(7,0,0)
	EndIf

	; Get stride
	; OpenGL uses bottom-up bitmaps ( stride < 0 )
	; If the image bitmap is top-down ( stride > 0 ) we have to rotate the image
	Local $iStride = DllStructGetData( $tBitmapData, "Stride" )
	If $iStride > 0 Then
		Local Enum _
			$RotateNoneFlipNone = 0 , _
			$Rotate90FlipNone = 1 , _
			$Rotate180FlipNone = 2 , _
			$Rotate270FlipNone = 3 , _
			$RotateNoneFlipX = 4 , _
			$Rotate90FlipX = 5 , _
			$Rotate180FlipX = 6 , _
			$Rotate270FlipX = 7 , _
			$RotateNoneFlipY = 6 , _
			$Rotate90FlipY = 7 , _
			$Rotate180FlipY = 4 , _
			$Rotate270FlipY = 5 , _
			$RotateNoneFlipXY = 6 , _
			$Rotate90FlipXY = 7 , _
			$Rotate180FlipXY = 0 , _
			$Rotate270FlipXY = 1

		; Rotate the image
		DllCall( $ghGDIPDll, "int", "GdipImageRotateFlip", "hwnd", $hBitmapObj, "long", $Rotate180FlipX )
		If @error Then
			_GDIPlus_Shutdown()
			PrintError( "Function LoadGDIimage()" & @CRLF & _
				"ERROR: Could not rotate image """ & $imgFile & """." )
			Return SetError(8,0,0)
		EndIf
	EndIf

	; Create byte structure
	Local $pScan0 = DllStructGetData( $tBitmapData, "Scan0" )
	Local $image = DllStructCreate( "byte[" & $size & "]", $pScan0 )
	Local $pImage = DllStructGetPtr( $image )

	Local $iRet = 1

	; Create texture
	If Not CreateTexture( $texture, $width, $height, $format, $pImage, $fTexGlu ) Then
		PrintTextureError( "LoadGDIimage()", @error, @extended )
		$iRet = 0
	EndIf

	; Clean up resources
	_GDIPlus_BitmapUnlockBits( $hBitmapObj, $tBitmapData )
	_GDIPlus_BitmapDispose( $hBitmapObj )

	; Close GDIPlus
	_GDIPlus_Shutdown()

	Return $iRet
EndFunc


; -------------------- LoadTGAimage() --------------------

Global Const $tagTGAheader = _
	"byte idLength;" & _
	"byte ColorMapType;" & _
	"byte DataTypeCode;" & _
	"byte ColorMapOrigin[2];" & _
	"byte ColorMapLength[2];" & _
	"byte ColorMapDepth;" & _
	"byte xOrigin[2];" & _
	"byte yOrigin[2];" & _
	"byte Width[2];" & _
	"byte Height[2];" & _
	"byte BitsPerPixel;" & _
	"byte ImageDescriptor"

; Reads a TGA ( uncompressed or RLE-encoded ) image into a byte structure
; Calls CreateTexture() to load the byte structure into a texture
Func LoadTGAimage( ByRef $texture, $imgFile, ByRef $width, ByRef $height, $fTexGlu = 0 )

	; File format
	Local $sFormat = StringRight( $imgFile, StringLen( $imgFile ) - StringInStr( $imgFile, ".", 0, -1 ) )
	Switch $sFormat
		Case "tga"
		Case Else
			PrintError( "Function LoadTGAimage()" & @CRLF & _
				"ERROR: """ & $sFormat & """ is not a legal image format." & @CRLF & _
				"Legal formats: TGA ( uncompressed or RLE-encoded )" )
			Return SetError(1,0,0)
	EndSwitch

	; File exists?
	If Not FileExists( $imgFile ) Then
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: File """ & $imgFile & """ does not exists." )
		Return SetError(2,0,0)
	EndIf

	; Open file
	Local $hImgFile = _WinAPI_CreateFile( $imgFile, 2, 2 )
	If Not $hImgFile Then
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: Unable to open """ & $imgFile & """ for reading." )
		Return SetError(3,0,0)
	EndIf

	; Read TGA header
	Local $tTGAheader = DllStructCreate( $tagTGAheader )
	Local $tTGAlength = DllStructGetSize( $tTGAheader ), $iBytes
	_WinAPI_ReadFile( $hImgFile, DllStructGetPtr( $tTGAheader ), $tTGAlength, $iBytes )
	If $iBytes <> $tTGAlength Then
		_WinAPI_CloseHandle( $hImgFile )
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: """ & $imgFile & """ has incomplete TGA header." )
		Return SetError(4,0,0)
	EndIf

	; Get data type code
	;  2: Uncompressed RGB TGA image
	; 10: Runlength encoded (RLE) RGB TGA image
	Local $fDataTypeCode = Number( DllStructGetData( $tTGAheader, "DataTypeCode" ) )
	If $fDataTypeCode <> 2 And $fDataTypeCode <> 10 Then
		_WinAPI_CloseHandle( $hImgFile )
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: """ & $imgFile & """ is not an uncompressed or RLE-encoded TGA image." )
		Return SetError(5,0,0)
	EndIf

	; Bytes per pixel
	Local $iBytesPerPixel = Number( DllStructGetData( $tTGAheader, "BitsPerPixel" ) ) / 8, $format
	If $iBytesPerPixel = 3 Then
		$format = $GL_BGR
	ElseIf $iBytesPerPixel = 4 Then
		$format = $GL_BGRA
	Else
		_WinAPI_CloseHandle( $hImgFile )
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: """ & $imgFile & """ is not a 24/32-bit TGA image." )
		Return SetError(6,0,0)
	EndIf

	; TGA ID string
	Local $idLength = Number( DllStructGetData( $tTGAheader, "idLength" ) )
	If $idLength Then
		Local $tBuffer = DllStructCreate( "byte[" & $idLength & "]" )
		_WinAPI_ReadFile( $hImgFile, DllStructGetPtr( $tBuffer ), $idLength, $iBytes )
		If $iBytes <> $idLength Then
			_WinAPI_CloseHandle( $hImgFile )
			PrintError( "Function LoadTGAimage()" & @CRLF & _
				"ERROR: """ & $imgFile & """ has incomplete TGA ID string." )
			Return SetError(7,0,0)
		EndIf
	EndIf

	; TGA Color map
	Local $ColorMapLength = Number( DllStructGetData( $tTGAheader, "ColorMapLength" ) )
	Local $ColorMapDepth = Number( DllStructGetData( $tTGAheader, "ColorMapDepth" ) )
	Local $ColorMapSize = $ColorMapLength * $ColorMapDepth / 8
	If $ColorMapSize Then
		Local $tBuffer = DllStructCreate( "byte[" & $ColorMapSize & "]" )
		_WinAPI_ReadFile( $hImgFile, DllStructGetPtr( $tBuffer ), $ColorMapSize, $iBytes )
		If $iBytes <> $ColorMapSize Then
			_WinAPI_CloseHandle( $hImgFile )
			PrintError( "Function LoadTGAimage()" & @CRLF & _
				"ERROR: """ & $imgFile & """ has incomplete TGA color map." )
			Return SetError(8,0,0)
		EndIf
	EndIf

	; Create byte structure
	$width  = Number( DllStructGetData( $tTGAheader, "Width" ) )
	$height = Number( DllStructGetData( $tTGAheader, "Height" ) )
	Local $iPixelsSize = $width * $height * $iBytesPerPixel
	Local $tImage = DllStructCreate( "byte[" & $iPixelsSize & "]" )
	Local $pImage = DllStructGetPtr( $tImage )

	_WinAPI_ReadFile( $hImgFile, $pImage, $iPixelsSize, $iBytes )
	_WinAPI_CloseHandle( $hImgFile )

	; Uncompressed TGA file
	If $fDataTypeCode = 2 And $iBytes <> $iPixelsSize Then
		PrintError( "Function LoadTGAimage()" & @CRLF & _
			"ERROR: """ & $imgFile & """ (uncompressed) has incomplete TGA image." )
		Return SetError(9,0,0)

	; Runlength encoded (RLE) TGA file
	; ( Expects $iBytes < $iPixelsSize because of RLE-encoding )
	ElseIf $fDataTypeCode = 10 Then
		Local $n = 1, $j, $iBytes = 0, $bytes, $sBytes, $sPixels, $typeCount
		While $iBytes < $iPixelsSize
			$typeCount = DllStructGetData( $tImage, 1, $n )
			$j = BitAND( $typeCount, 0x7F ) + 1
			If BitAND( $typeCount, 0x80 ) Then
				; RLE chunk, the next $iBytesPerPixel ( 3 or 4 ) bytes are repeated $j times
				$bytes = DllStructCreate( "byte[" & $iBytesPerPixel & "]", $pImage + $n )
				$sBytes = BinaryToString( DllStructGetData( $bytes, 1 ) )
				For $i = 1 To $j
					$sPixels &= $sBytes
				Next
				$n += 1 + $iBytesPerPixel
			Else
				; Normal chunk, read the next $iBytesPerPixel * $j bytes
				$bytes = DllStructCreate( "byte[" & $iBytesPerPixel * $j & "]", $pImage + $n )
				$sPixels &= BinaryToString( DllStructGetData( $bytes, 1 ) )
				$n += 1 + $iBytesPerPixel * $j
			EndIf
			$iBytes += $iBytesPerPixel * $j
		WEnd
		If $iBytes <> $iPixelsSize Then
			PrintError( "Function LoadTGAimage()" & @CRLF & _
				"ERROR: """ & $imgFile & """ (RLE-encoded) has incomplete TGA image." )
			Return SetError(10,0,0)
		EndIf
		DllStructSetData( $tImage, 1, $sPixels )

	EndIf

	; Create texture
	If Not CreateTexture( $texture, $width, $height, $format, $pImage, $fTexGlu ) Then
		PrintTextureError( "LoadTGAimage()", @error, @extended )
		Return 0
	Else
		Return 1
	EndIf

EndFunc


; -------------------- LoadRGBimage() --------------------

; Reads a RGB and RGBA image into a byte structure
; Calls CreateTexture() to load the byte structure into a texture
Func LoadRGBimage( ByRef $texture, $imgFile, $width, $height, $fTexGlu = 0 )
	Local $bytes, $format, $size, $filesize

	; File format
	Local $sFormat = StringRight( $imgFile, StringLen( $imgFile ) - StringInStr( $imgFile, ".", 0, -1 ) )
	Switch $sFormat
		Case "rgb"
			$bytes = 3
			$format = $GL_RGB
		Case "rgba"
			$bytes = 4
			$format = $GL_RGBA
		Case Else
			PrintError( "Function LoadRGBimage()" & @CRLF & _
				"ERROR: """ & $sFormat & """ is not a legal image format." & @CRLF & _
				"Legal formats: RGB, RGBA" )
			Return SetError(1,0,0)
	EndSwitch

	; File exists?
	If Not FileExists( $imgFile ) Then
		PrintError( "Function LoadRGBimage()" & @CRLF & _
			"ERROR: File """ & $imgFile & """ does not exists." )
		Return SetError(2,0,0)
	EndIf

	; File size
	$size = $width * $height * $bytes
	$filesize = FileGetSize( $imgFile )
	If $filesize <> $size Then
		PrintError( "Function LoadRGBimage()" & @CRLF & _
			"ERROR: File size = " & $filesize & " for """ & $imgFile & """" & @CRLF & _
			"does not match $width * $height * $bytes = " & $size )
		Return SetError(3,0,0)
	EndIf

	; Read file
	Local $hImage = FileOpen( $imgFile, 16 )
	Local $image = DllStructCreate( "byte[" & $size & "]" )
	DllStructSetData( $image, 1, BinaryToString( FileRead( $hImage ) ) )
	Local $pImage = DllStructGetPtr( $image )
	FileClose( $hImage )

	; Create texture
	If Not CreateTexture( $texture, $width, $height, $format, $pImage, $fTexGlu ) Then
		PrintTextureError( "LoadRGBimage()", @error, @extended )
		Return 0
	Else
		Return 1
	EndIf
EndFunc


; -------------------- CreateTexture() --------------------

; Loads a byte structure into a texture and creates mipmaps
Func CreateTexture( ByRef $texture, $width, $height, $format, $pImage, $fTexGlu = 0 )

	; If the video card supports OpenGL 2.1 or higher glTexImage2D() and glGenerateMipmap() will be used to load images into
	; textures and create mipmaps.

	; If the supported version is 2.0 or lower gluBuild2DMipmaps() will be used to load images into textures and create mipmaps.

	; Because gluBuild2DMipmaps() is a very widely used function, you can specify $fTexGlu = 1 to force usage of the GLU function.

	Local Static $fTexMethod = 0
	Local Static $fMaxAnisotrophy = 0.0

	If Not $fTexGlu And Not $fTexMethod Then

		; Only do these calculations once

		Local $sVersion = StringLeft( glGetString( $GL_VERSION ), 3 )
		Local $iMajor = Number( StringLeft( $sVersion, 1 ) )   ; Major version number supported by video card
		Local $iMinor = Number( StringRight( $sVersion, 1 ) )  ; Minor version number supported by video card

		Select

			Case $iMajor = 2 And $iMinor = 1
				; glGenerateMipmap() can be used if one of the GL_ARB_framebuffer_object or GL_EXT_framebuffer_object extensions exists

				Local $sExtensions = glGetString( $GL_EXTENSIONS )
				Local $aExtensions = StringSplit( $sExtensions, " " )
				Local $iExtensions = $aExtensions[0]

				$fTexMethod = 1

				; Check for GL_ARB_framebuffer_object or GL_EXT_framebuffer_object
				For $i = 1 To $iExtensions
					If $aExtensions[$i] = "GL_ARB_framebuffer_object" Or _
						 $aExtensions[$i] = "GL_EXT_framebuffer_object" Then
								$fTexMethod = 2
								ExitLoop
					EndIf
				Next

			Case $iMajor >= 3
				; glGenerateMipmap() OK
				$fTexMethod = 2

			Case Else
				; 1.x, 2.0: Use gluBuild2DMipmaps()
				$fTexMethod = 1

		EndSelect

	EndIf

	Local $iOpenGLerror = glGetError()

	glEnable( $GL_TEXTURE_2D )
	glGenTextures( 1, $texture )
	glBindTexture( $GL_TEXTURE_2D, $texture )
	$iOpenGLerror = glGetError()
	If $iOpenGLerror <> $GL_NO_ERROR Then _
		Return SetError( 1, $iOpenGLerror, 0 )

	If Not $fTexGlu Then

		Switch $fTexMethod

			Case 1 ; Old method
				; http://www.opengl.org/sdk/docs/man2/xhtml/gluBuild2DMipmaps.xml
				gluBuild2DMipmaps( $GL_TEXTURE_2D, $GL_RGBA8, $width, $height, $format, $GL_UNSIGNED_BYTE, $pImage )
				$iOpenGLerror = glGetError()
				If $iOpenGLerror <> $GL_NO_ERROR Then _
					Return SetError( 2, $iOpenGLerror, 0 )

			Case 2 ; Newer method
				glTexImage2D( $GL_TEXTURE_2D, 0, $GL_RGBA8, $width, $height, 0, $format, $GL_UNSIGNED_BYTE, $pImage )
				glGenerateMipmap( $GL_TEXTURE_2D )
				$iOpenGLerror = glGetError()
				If $iOpenGLerror <> $GL_NO_ERROR Then _
					Return SetError( 3, $iOpenGLerror, 0 )

		EndSwitch

	Else

		; http://www.opengl.org/sdk/docs/man2/xhtml/gluBuild2DMipmaps.xml
		gluBuild2DMipmaps( $GL_TEXTURE_2D, $GL_RGBA8, $width, $height, $format, $GL_UNSIGNED_BYTE, $pImage )
		$iOpenGLerror = glGetError()
		If $iOpenGLerror <> $GL_NO_ERROR Then _
			Return SetError( 2, $iOpenGLerror, 0 )

	EndIf

	glTexParameteri( $GL_TEXTURE_2D, $GL_TEXTURE_WRAP_S, $GL_REPEAT )
	glTexParameteri( $GL_TEXTURE_2D, $GL_TEXTURE_WRAP_T, $GL_REPEAT )
	glTexParameteri( $GL_TEXTURE_2D, $GL_TEXTURE_MAG_FILTER, $GL_LINEAR )
	glTexParameteri( $GL_TEXTURE_2D, $GL_TEXTURE_MIN_FILTER, $GL_LINEAR_MIPMAP_LINEAR )

	If $fMaxAnisotrophy = 0.0 Then
		If CheckOpenGLextension( "GL_EXT_texture_filter_anisotropic" ) Then
			glGetFloatv( $GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, $fMaxAnisotrophy )
			If $fMaxAnisotrophy > 1.0 Then
				;MsgBox( 0, "CreateTexture: $fMaxAnisotrophy", $fMaxAnisotrophy )
				glTexParameterf( $GL_TEXTURE_2D, $GL_TEXTURE_MAX_ANISOTROPY_EXT, $fMaxAnisotrophy )
			Else
				$fMaxAnisotrophy = 1.0
			EndIf
		Else
			$fMaxAnisotrophy = 1.0
		EndIf
	EndIf

	Return 1

EndFunc

Func PrintTextureError( $sFunction, $iError, $iExtended )
	Local $sOpenGLfunc
	Switch $iError
		Case 1
			$sOpenGLfunc = "glGenTextures() or glBindTexture()"
		Case 2
			$sOpenGLfunc = "gluBuild2DMipmaps()"
		Case 3
			$sOpenGLfunc = "glTexImage2D() or glGenerateMipmap()"
	EndSwitch
	PrintOpenGLerror( $sFunction, $sOpenGLfunc, $iExtended )
EndFunc
